import scipy as sp
import scipy.stats as st


def _qqplot_bar(M=1000000, alphaLevel=0.05, distr='log10'):
    """calculate theoretical expectations for qqplot"""
    mRange = 10**(sp.arange(sp.log10(0.5), sp.log10(M - 0.5) +
                            0.1, 0.1))  # should be exp or 10**?
    numPts = len(mRange)
    betaalphaLevel = sp.zeros(numPts)  # down in the plot
    betaOneMinusalphaLevel = sp.zeros(numPts)  # up in the plot
    betaInvHalf = sp.zeros(numPts)
    for n in range(numPts):
        m = mRange[n]  # numPLessThanThresh=m
        betaInvHalf[n] = st.beta.ppf(0.5, m, M - m)
        betaalphaLevel[n] = st.beta.ppf(alphaLevel, m, M - m)
        betaOneMinusalphaLevel[n] = st.beta.ppf(1 - alphaLevel, m, M - m)
    betaDown = betaInvHalf - betaalphaLevel
    betaUp = betaOneMinusalphaLevel - betaInvHalf

    theoreticalPvals = mRange / M
    return betaUp, betaDown, theoreticalPvals


def qqplot(pv, label='unknown', distr='log10', alphaLevel=0.05, ax=None,
           color=None):
    r"""Produces a Quantile-Quantile plot of the observed P value
        distribution against the theoretical one under the null.

    Parameters
    ----------

    pv : array_like
        P-values.
    distr : {'log10', 'chi2'}
        Scale of the distribution. If 'log10' is specified, the distribution
        of the -log10 P values is considered.
        If the distribution of the corresponding chi2-distributed test
        statistics is considered. Defaults to 'log10'.
    alphaLevel : float
        Significance bound.
    ax : :class:`matplotlib.axes.AxesSubplot`
        The target handle for this figure. If None, the current axes is set.

    Returns
    -------
    :class:`matplotlib.axes.AxesSubplot`
        Axes.

    Examples
    --------

    .. plot::

        from limix.plot import qqplot
        from numpy.random import RandomState
        from matplotlib import pyplot as plt
        random = RandomState(1)

        pv = random.rand(10000)

        fig = plt.figure(1, figsize=(5,5))
        plt.subplot(111)
        qqplot(pv)
        plt.tight_layout()
        plt.show()
    """
    import matplotlib.pylab as plt
    if ax is None:
        ax = plt.gca()

    shape_ok = (len(pv.shape) == 1) or (
        (len(pv.shape) == 2) and pv.shape[1] == 1)
    assert shape_ok, 'qqplot requires a 1D array of p-values'

    tests = pv.shape[0]
    pnull = (0.5 + sp.arange(tests)) / tests
    # pnull = np.sort(np.random.uniform(size = tests))
    Ipv = sp.argsort(pv)

    if distr == 'chi2':
        qnull = sp.stats.chi2.isf(pnull, 1)
        qemp = (sp.stats.chi2.isf(pv[Ipv], 1))
        xl = 'LOD scores'
        yl = '$\chi^2$ quantiles'

    if distr == 'log10':
        qnull = -sp.log10(pnull)
        qemp = -sp.log10(pv[Ipv])

        xl = '-log10(P) observed'
        yl = '-log10(P) expected'

    plt.plot(qnull, qemp, '.', color=color, label=label)
    # plt.plot([0,qemp.m0x()], [0,qemp.max()],'r')
    plt.plot([0, qnull.max()], [0, qnull.max()], 'r')
    plt.ylabel(xl)
    plt.xlabel(yl)
    if alphaLevel is not None:
        if distr == 'log10':
            betaUp, betaDown, theoreticalPvals = _qqplot_bar(
                M=tests, alphaLevel=alphaLevel, distr=distr)
            lower = -sp.log10(theoreticalPvals - betaDown)
            upper = -sp.log10(theoreticalPvals + betaUp)
            plt.fill_between(-sp.log10(theoreticalPvals),
                             lower, upper, color='grey', alpha=0.5)
            # plt.plot(-sp.log10(theoreticalPvals),lower,'g-.')
            # plt.plot(-sp.log10(theoreticalPvals),upper,'g-.')

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    return ax



r"""
***********************
Heritability estimation
***********************

- :func:`.estimate`

Heritability
^^^^^^^^^^^^

.. autofunction:: limix.heritability.estimate

"""

from .estimate import estimate

__all__ = ['estimate']



r"""
*****************
Utility functions
*****************

- :func:`.sets_from_bim`
- :func:`.annotate_sets`
- :func:`.estCumPos`
- :func:`.unique_variants`
- :class:`.TemporaryDirectory`
- :func:`.urlretrieve`

Public interface
^^^^^^^^^^^^^^^^
"""

from .set_utils import sets_from_bim
from .set_utils import annotate_sets
from .geno_utils import estCumPos
from .geno_utils import unique_variants
from .temp import TemporaryDirectory
from .url import urlretrieve
from . import preprocess

__all__ = ['sets_from_bim',
           'annotate_sets',
           'estCumPos',
           'unique_variants',
           'TemporaryDirectory',
           'urlretrieve']



import copy

import pandas as pd
import scipy as sp
from sklearn.preprocessing import Imputer

from limix.io import read_plink


class BedReader():
    r"""
    Class to read and make queries on plink binary files.

    Parameters
    ----------
    prefix : str
        Path prefix to the set of PLINK files.

    Examples
    --------

    Basics

    .. doctest::

        >>> from limix.data import BedReader
        >>> from limix.data import build_geno_query
        >>> from pandas_plink import example_file_prefix
        >>>
        >>> reader = BedReader(example_file_prefix())
        >>>
        >>> print(reader.getSnpInfo().head())
          chrom         snp   cm    pos a0 a1  i
        0     1  rs10399749  0.0  45162  G  C  0
        1     1   rs2949420  0.0  45257  C  T  1
        2     1   rs2949421  0.0  45413  0  0  2
        3     1   rs2691310  0.0  46844  A  T  3
        4     1   rs4030303  0.0  72434  0  G  4

    Query and load genotype values into memory:

    .. doctest::

        >>> # build genotype query
        >>> gquery = build_geno_query(idx_start=4,
        ...                           idx_end=10,
        ...                           pos_start=45200,
        ...                           pos_end=80000,
        ...                           chrom=1)
        >>>
        >>> # apply geno query and impute
        >>> X, snpinfo = reader.getGenotypes(gquery,
        ...                                  impute=True,
        ...                                  return_snpinfo=True)
        >>>
        >>> print(snpinfo)
          chrom        snp   cm    pos a0 a1  i
        0     1  rs4030303  0.0  72434  0  G  4
        1     1  rs4030300  0.0  72515  0  C  5
        2     1  rs3855952  0.0  77689  G  A  6
        3     1   rs940550  0.0  78032  0  T  7
        >>>
        >>> print(X) # doctest: +SKIP
        [[2. 2. 2. 2.]
         [2. 2. 1. 2.]
         [2. 2. 0. 2.]]

    Lazy subsetting using queries:

    .. doctest::

        >>> reader_sub = reader.subset_snps(gquery)
        >>>
        >>> print(reader_sub.getSnpInfo().head())
          chrom        snp   cm    pos a0 a1  i
        0     1  rs4030303  0.0  72434  0  G  0
        1     1  rs4030300  0.0  72515  0  C  1
        2     1  rs3855952  0.0  77689  G  A  2
        3     1   rs940550  0.0  78032  0  T  3
        >>>
        >>> # only when using getGenotypes, the genotypes are loaded
        >>> print( reader_sub.getGenotypes( impute=True ) ) # doctest: +SKIP
        [[2. 2. 2. 2.]
         [2. 2. 1. 2.]
         [2. 2. 0. 2.]]

    You can do it in place as well:

    .. doctest::

        >>> query1 = build_geno_query(pos_start=72500, pos_end=78000)
        >>>
        >>> reader_sub.subset_snps(query1, inplace=True)
        >>>
        >>> print(reader_sub.getSnpInfo())
          chrom        snp   cm    pos a0 a1  i
        0     1  rs4030300  0.0  72515  0  C  0
        1     1  rs3855952  0.0  77689  G  A  1

    and you can even iterate on genotypes to enable
    low-memory genome-wide analyses.

    .. doctest::

        >>> from limix.data import GIter
        >>>
        >>> for gr in GIter(reader, batch_size=2):
        ...     print(gr.getGenotypes().shape)
        (3, 2)
        (3, 2)
        (3, 2)
        (3, 2)
        (3, 2)

    Have fun!

    """

    def __init__(self, prefix):
        self._prefix = prefix
        self._load()
        self._init_imputer()

    def _load(self):
        (bim, fam, bed) = read_plink(self._prefix, verbose=False)
        self._snpinfo = bim
        self._ind_info = fam
        self._geno = bed

    def _init_imputer(self):
        self._imputer = Imputer(
            missing_values=3., strategy='mean', axis=0, copy=False)

    def __str__(self):
        rv = '<' + str(self.__class__)
        rv += ' instance at '
        rv += hex(id(self)) + '>\n'
        rv += 'File: ' + self._prefix + '\n'
        rv += 'Dims: %d inds, %d snps' % (self._geno.shape[1],
                                          self._geno.shape[0])
        return rv

    def getSnpInfo(self):
        r"""
        Return pandas dataframe with all variant info.
        """
        return self._snpinfo

    def subset_snps(self, query=None, inplace=False):
        r""" Builds a new bed reader with filtered variants.

        Parameters
        ----------
        query : str
            pandas query on the bim file.
            The default value is None.
        inplace : bool
            If True, the operation is done in place.
            Default is False.

        Returns
        -------
            R : :class:`limix.BedReader`
                Bed reader with filtered variants
                (if inplace is False).
        """
        # query
        geno, snpinfo = self._query(query)
        snpinfo = snpinfo.assign(
            i=pd.Series(sp.arange(snpinfo.shape[0]), index=snpinfo.index))

        if inplace:
            # replace
            self._geno = geno
            self._snpinfo = snpinfo
        else:
            # copy (note the first copy is not deep)
            R = copy.copy(self)
            R._ind_info = copy.copy(self._ind_info)
            R._geno = geno
            R._snpinfo = snpinfo
            return R

    def getGenotypes(self,
                     query=None,
                     impute=False,
                     standardize=False,
                     return_snpinfo=False):
        r""" Query and Load genotype data.

        Parameters
        ----------
        query : str
            pandas query on the bim file.
            The default is None.
        impute : bool, optional
            list of chromosomes.
            If True,
            the missing values in the bed file are mean
            imputed (variant-by-variant).
            If standardize is True, the default value of
            impute is True, otherwise is False.
        standardize : bool, optional
            If True, the genotype values are standardizes.
            The default value is False.
        return_snpinfo : bool, optional
            If True, returns genotype info
            By default is False.

        Returns
        -------
            X : ndarray
                (`N`, `S`) ndarray of queried genotype values
                for `N` individuals and `S` variants.
            snpinfo : :class:`pandas.DataFrame`
                dataframe with genotype info.
                Returned only if ``return_snpinfo=True``.
        """
        if standardize:
            impute = True

        # query
        geno, snpinfo = self._query(query)

        # compute
        X = geno.compute().T

        # impute and standardize
        if impute:
            X = self._imputer.fit_transform(X)

        if standardize:
            X = X.astype(float)
            X -= X.mean(0)
            X /= X.std(0)

        if return_snpinfo:
            return X, snpinfo
        else:
            return X

    def _query(self, query):
        if query is None:
            return self._geno, self._snpinfo
        snpinfo = self._snpinfo.query(query)
        snpinfo.reset_index(inplace=True, drop=True)
        geno = self._geno[snpinfo.i.values, :]
        return geno, snpinfo



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
qtl.py contains wrappers around C++ Limix objects to streamline common tasks in GWAS.
"""

import time

import numpy as np
import scipy as sp
import scipy.stats as st

from limix.qtl.lmm import LMM
from limix.stats import qvalues


def qtl_test_lm(snps, pheno, covs=None, test='lrt', verbose=None):
    """
    Wrapper function for univariate single-variant association testing
    using a linear model.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        covs (ndarray, optional):
            (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        test ({'lrt', 'f'}, optional):
            test statistic.
            'lrt' for likelihood ratio test (default) or 'f' for F-test.
        verbose (bool, optional):
            if True, details such as runtime as displayed.

    Returns:
        :class:`limix.qtl.LMM`: LIMIX LMM object

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.qtl import qtl_test_lm
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, 1)
            >>>
            >>> lm = qtl_test_lm(snps, pheno)
            >>> print(lm.getPv()[:,:4]) # doctest: +SKIP
            [[0.8796 0.5065 0.5666 0.6016]]
    """
    lm = qtl_test_lmm(
        snps=snps, pheno=pheno, K=None, covs=covs, test=test, verbose=verbose)
    return lm


def qtl_test_lmm(snps,
                 pheno,
                 K=None,
                 covs=None,
                 test='lrt',
                 NumIntervalsDelta0=100,
                 NumIntervalsDeltaAlt=100,
                 searchDelta=False,
                 verbose=None):
    """
    Wrapper function for univariate single-variant association testing
    using a linear mixed model.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        K (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            If not provided, then standard linear regression is considered.
        covs (ndarray, optional):
            (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        test ({'lrt', 'f'}, optional):
            test statistic.
            'lrt' for likelihood ratio test (default) or 'f' for F-test.
        NumIntervalsDelta0 (int, optional):
            number of steps for delta optimization on the null model.
            By default ``NumIntervalsDelta0`` is 100.
        NumIntervalsDeltaAlt (int, optional):
            number of steps for delta optimization on the alternative model.
            Requires ``searchDelta=True`` to have an effect.
        searchDelta (bool, optional):
            if True, delta optimization on the alternative model is carried out.
            By default ``searchDelta`` is False.
        verbose (bool, optional):
            if True, details such as runtime as displayed.

    Returns:
        :class:`limix.qtl.LMM`: LIMIX LMM object

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import dot
            >>> from limix.qtl import qtl_test_lmm
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, 1)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>>
            >>> lmm = qtl_test_lmm(snps, pheno, kinship)
            >>> print(lmm.getPv()[:,:4]) # doctest: +SKIP
            [[0.8571 0.4668 0.5872 0.5589]]
    """
    lmm_ = LMM(
        snps=snps,
        pheno=pheno,
        K=K,
        covs=covs,
        test=test,
        NumIntervalsDelta0=NumIntervalsDelta0,
        NumIntervalsDeltaAlt=NumIntervalsDeltaAlt,
        searchDelta=searchDelta,
        verbose=verbose)
    return lmm_


def qtl_test_lmm_kronecker(snps,
                           phenos,
                           covs=None,
                           Acovs=None,
                           Asnps=None,
                           K1r=None,
                           K1c=None,
                           K2r=None,
                           K2c=None,
                           NumIntervalsDelta0=100,
                           NumIntervalsDeltaAlt=100,
                           searchDelta=False):
    r"""
    Wrapper function for single-variant multi-trait association testing class
    using linear mixed models.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        covs (ndarray or list of ndarrays, optional):
            sample design matrices of the different fixed effect terms.
            In the case of a single fixed effect term, ``covs`` can be specified
            as a (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            In this case ``Acovs`` should also be a (`K`, `P`) ndarray,
            where `K` is the rank of the trait design and
            `P` is the number of traits.
            For multiple terms, ``covs`` is a list of ndarray, where each
            element ``covs[i]`` has shape (`N`, `Di`).
            In this case, ``Acovs`` should also be list with the same number
            of  ndarray elements.
            Each element ``Acovs[i]`` has shape (`Ki`, `P`), where `P` is the
            number of analyzed phenotypes.
            By default covs is (`N`, `1`) ndarray of ones and
            ``Acovs[i]`` is a (`P`, `P`) identity ndarray.
        Acovs (ndarray or list of ndarrays, optional):
            trait design matrices of the different fixed effect terms.
            See `covs` for more info.
        Asnps (ndarray, optional):
            (`K`, `P`) trait design of rank `K` of the snp fixed effect on the
            `P` traits. By default ``Asnps`` is (`1`, `P`) ndrray of ones.
        K1r (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            If not provided, ``K1r = numpy.dot(snps, snps.T)`` is assumed.
        K1c (ndarray, optional):
            (`P`, `P`) ndarray of LMM-covariance/kinship coefficients.
            If either `K1c` or `K2c` is not provided, K1c is estimated by
            (restricted) maximum likelihood  from the model without the snp effect.
        K2r (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            By default ``K2r`` is a (`N`, `N`) identity ndarray.
        K2c (ndarray, optional):
            (`P`, `P`) ndarray of residual covariances.
            If either `K1c` or `K2c` is not provided, K1c is estimated by
            (restricted) maximum likelihood  from the model without the snp effect.
        NumIntervalsDelta0 (int, optional):
            number of steps for delta optimization on the null model.
            By default ``NumIntervalsDelta0`` is 100.
        NumIntervalsDeltaAlt (int, optional):
            number of steps for delta optimization on the alternative model.
            Requires ``searchDelta=True`` to have an effect.
        searchDelta (bool, optional):
            if True, delta optimization on the alternative model is carried out.
            By default ``searchDelta`` is False.

    Returns:
        (tuple): tuple containing:
            - **lmm** (*:class:`limix_legacy.deprecated.CKroneckerLMM()`*):
              CKroneckerLMM limix_legacy object
            - **pv** (*ndarray*):
              P-values for all SNPs from liklelihood ratio test

    Examples
    --------

        Example of multi-trait single-variant association testing using
        a linear mixed model and specifying a common effects.

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import dot, eye, ones
            >>> from limix.qtl import qtl_test_lmm_kronecker
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>> P = 2
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, P)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>> kinship+= 1e-4 * eye(N)
            >>>
            >>> #common effect test
            >>> Asnps = ones((1, P))
            >>> lmm, pv = qtl_test_lmm_kronecker(snps, pheno, K1r=kinship,
            ...                                 Asnps=Asnps)
            >>> beta = lmm.getBetaSNP()
            >>>
            >>> print(pv.shape)
            (1, 1000)
            >>> print(pv[:,:4]) # doctest: +SKIP
            [[0.8722 0.6509 0.3593 0.6816]]
            >>>
            >>> print(beta.shape)
            (1, 1000)
            >>> print(beta[:,:4]) # doctest: +SKIP
            [[0.0281 0.0758 0.1458 0.0701]]

        Example showing how to set an any effect test.
        For more information on effect designs
        see `limix_tutorials.`_

        .. _limix_tutorials.: http://nbviewer.jupyter.org/github/limix/limix-tutorials/blob/master/index.ipynb

        .. doctest::

            >>> #any effect test
            >>> Asnps = eye(P)
            >>> lmm, pv = qtl_test_lmm_kronecker(snps, pheno, K1r=kinship,
            ...                                 Asnps=Asnps)
            >>> beta = lmm.getBetaSNP()
            >>>
            >>> print(pv.shape)
            (1, 1000)
            >>> print(pv[:,:4]) # doctest: +SKIP
            [[0.9838 0.8664 0.3555 0.5244]]
            >>>
            >>> print(beta.shape)
            (2, 1000)
            >>> print(beta[:,:4]) # doctest: +SKIP
            [[ 0.0413  0.0312 -0.0177 -0.098 ]
             [ 0.0139  0.1234  0.3204  0.2495]]
    """
    import limix_legacy.deprecated
    import limix_legacy.deprecated as dlimix_legacy
    # 0. checks
    N = phenos.shape[0]
    P = phenos.shape[1]

    if K1r is None:
        K1r = np.dot(snps, snps.T)
    else:
        assert K1r.shape[0] == N, 'K1r: dimensions dismatch'
        assert K1r.shape[1] == N, 'K1r: dimensions dismatch'

    if K2r is None:
        K2r = np.eye(N)
    else:
        assert K2r.shape[0] == N, 'K2r: dimensions dismatch'
        assert K2r.shape[1] == N, 'K2r: dimensions dismatch'

    covs, Acovs = _updateKronCovs(covs, Acovs, N, P)

    # Asnps can be several designs
    if Asnps is None:
        Asnps = [np.ones([1, P])]
    if (not isinstance(Asnps, list)):
        Asnps = [Asnps]
    assert len(Asnps) > 0, "need at least one Snp design matrix"

    # one row per column design matrix
    pv = np.zeros((len(Asnps), snps.shape[1]))

    # 1. run GP model to infer suitable covariance structure
    if K1c is None or K2c is None:
        vc = _estimateKronCovariances(
            phenos=phenos,
            K1r=K1r,
            K2r=K2r,
            K1c=K1c,
            K2c=K2c,
            covs=covs,
            Acovs=Acovs,
            verbose=False)
        K1c = vc.getTraitCovar(0)
        K2c = vc.getTraitCovar(1)
    else:
        assert K1c.shape[0] == P, 'K1c: dimensions dismatch'
        assert K1c.shape[1] == P, 'K1c: dimensions dismatch'
        assert K2c.shape[0] == P, 'K2c: dimensions dismatch'
        assert K2c.shape[1] == P, 'K2c: dimensions dismatch'

    # 2. run kroneckerLMM
    lmm = limix_legacy.deprecated.CKroneckerLMM()
    lmm.setK1r(K1r)
    lmm.setK1c(K1c)
    lmm.setK2r(K2r)
    lmm.setK2c(K2c)
    lmm.setSNPs(snps)
    # add covariates
    for ic in range(len(Acovs)):
        lmm.addCovariates(covs[ic], Acovs[ic])
    lmm.setPheno(phenos)

    # delta serch on alt. model?
    if searchDelta:
        lmm.setNumIntervalsAlt(NumIntervalsDeltaAlt)
    else:
        lmm.setNumIntervalsAlt(0)
    lmm.setNumIntervals0(NumIntervalsDelta0)

    for iA in range(len(Asnps)):
        # add SNP design
        lmm.setSNPcoldesign(Asnps[iA])
        lmm.process()
        pv[iA, :] = lmm.getPv()[0]
    return lmm, pv


def qtl_test_interaction_lmm_kronecker(snps,
                                       phenos,
                                       covs=None,
                                       Acovs=None,
                                       Asnps1=None,
                                       Asnps0=None,
                                       K1r=None,
                                       K1c=None,
                                       K2r=None,
                                       K2c=None,
                                       NumIntervalsDelta0=100,
                                       NumIntervalsDeltaAlt=100,
                                       searchDelta=False,
                                       return_lmm=False):
    r"""
    Wrapper function for single-variant multi-trait interaction test

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        covs (ndarray or list of ndarrays, optional):
            sample design matrices of the different fixed effect terms.
            See `limix.qtl.qtl_test_lmm_kronecker` for more info.
        Acovs (ndarray or list of ndarrays, optional):
            trait design matrices of the different fixed effect terms.
            See `limix.qtl.qtl_test_lmm_kronecker` for more info.
        Asnps1 (ndarray or list of ndarray, optional):
            (`K`, `P`) trait design of rank `K` of the snp fixed effect
            on the `P` traits for the alternative model.
            ``Asnps1`` is tested against the null trait design ``Asnps0``.
            Multiple designs can be tested vs the null trait design
            If a list of ndarray trait designs is specified for ``Asnps1``,
            the multiple designs are tested vs the null trait design
            ``Asnps0`` one-by-one.
            By default ``Asnps1`` is a (`P`, `P`) identity ndarray.
        Asnps0 (ndarray, optional):
            (`K0`, `P`) trait design of rank `K` of the snp fixed effect
            on the `P` traits for the null model.
            By default ``Asnps0`` is (`1`, `P`) ndrray of ones.
        K1r (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            If not provided, ``K1r = numpy.dot(snps, snps.T)`` is assumed.
        K1c (ndarray, optional):
            (`P`, `P`) ndarray of LMM-covariance/kinship coefficients.
            If either `K1c` or `K2c` is not provided, K1c is estimated by
            (restricted) maximum likelihood  from the model without the snp effect.
        K2r (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            By default ``K2r`` is a (`N`, `N`) identity ndarray.
        K2c (ndarray, optional):
            (`P`, `P`) ndarray of residual covariances.
            If either `K1c` or `K2c` is not provided, K1c is estimated by
            (restricted) maximum likelihood  from the model without the snp effect.
        NumIntervalsDelta0 (int, optional):
            number of steps for delta optimization on the null model.
            By default ``NumIntervalsDelta0`` is 100.
        NumIntervalsDeltaAlt (int, optional):
            number of steps for delta optimization on the alternative model.
            Requires ``searchDelta=True`` to have an effect.
        searchDelta (bool, optional):
            if True, delta optimization on the alternative model is carried out.
            By default ``searchDelta`` is False.

    Returns:
        (tuple): tuple containing:
            - **pv** (*ndarray*):
              P-values of the interaction test (``Asnps1`` vs ``Asnps0``)
              If ``Asnps1`` is a list of trait designs,
              `pv` contains the P values for the different trait designs (as rows).
            - **pv0** (*ndarray*):
              P-values of the association test ``Asnps0`` vs ``0``.
            - **pvAlt** (*ndarray*):
              P-values of the association test ``Asnps1`` vs ``0``.
              If ``Asnps1`` is a list of trait designs,
              `pvAlt` contains the P values for the different
              trait designs (as rows).

    Examples
    --------

        Example of how to perform a specific, common and any effect test.
        For more information see `limix_tutorials.`_

        .. _limix_tutorials.: http://nbviewer.jupyter.org/github/limix/limix-tutorials/blob/master/index.ipynb

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import dot, eye, ones
            >>> from limix.qtl import qtl_test_interaction_lmm_kronecker
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>> P = 2
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, P)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>> kinship+= 1e-4 * eye(N)
            >>>
            >>> #interaction effect test
            >>> Asnps0 = ones((1,P))
            >>> Asnps1 = eye(P)
            >>> pv, pv0, pvAlt = qtl_test_interaction_lmm_kronecker(snps, pheno,
            ...                                                     K1r=kinship,
            ...                                                     Asnps0=Asnps0,
            ...                                                     Asnps1=Asnps1)
            >>> #interaction P value
            >>> print(pv.shape)
            (1, 1000)
            >>> print(pv[:,:4]) # doctest: +SKIP
            [[0.9347 0.7744 0.2678 0.2893]]
            >>>
            >>> #common effect P value
            >>> print(pv0.shape)
            (1, 1000)
            >>> print(pv0[:,:4]) # doctest: +SKIP
            [[0.8722 0.6509 0.3593 0.6816]]
            >>>
            >>> #any effect P value
            >>> print(pvAlt.shape)
            (1, 1000)
            >>> print(pvAlt[:,:4]) # doctest: +SKIP
            [[0.9838 0.8664 0.3555 0.5244]]
    """
    import limix_legacy.deprecated
    import limix_legacy.deprecated as dlimix_legacy
    S = snps.shape[1]
    # 0. checks
    N = phenos.shape[0]
    P = phenos.shape[1]

    if K1r is None:
        K1r = np.dot(snps, snps.T)
    else:
        assert K1r.shape[0] == N, 'K1r: dimensions dismatch'
        assert K1r.shape[1] == N, 'K1r: dimensions dismatch'

    if K2r is None:
        K2r = np.eye(N)
    else:
        assert K2r.shape[0] == N, 'K2r: dimensions dismatch'
        assert K2r.shape[1] == N, 'K2r: dimensions dismatch'

    covs, Acovs = _updateKronCovs(covs, Acovs, N, P)

    # Asnps can be several designs
    if (Asnps0 is None):
        Asnps0 = [np.ones([1, P])]
    if Asnps1 is None:
        Asnps1 = [np.eye([P])]
    if (not isinstance(Asnps0, list)):
        Asnps0 = [Asnps0]
    if (not isinstance(Asnps1, list)):
        Asnps1 = [Asnps1]
    assert (len(Asnps0) == 1) and (
        len(Asnps1) >
        0), "need at least one Snp design matrix for null and alt model"

    # one row per column design matrix
    pv = np.zeros((len(Asnps1), snps.shape[1]))
    lrt = np.zeros((len(Asnps1), snps.shape[1]))
    pvAlt = np.zeros((len(Asnps1), snps.shape[1]))
    lrtAlt = np.zeros((len(Asnps1), snps.shape[1]))

    # 1. run GP model to infer suitable covariance structure
    if K1c is None or K2c is None:
        vc = _estimateKronCovariances(
            phenos=phenos,
            K1r=K1r,
            K2r=K2r,
            K1c=K1c,
            K2c=K2c,
            covs=covs,
            Acovs=Acovs,
            trait_covar_type='freeform',
            verbose=False)
        K1c = vc.getTraitCovar(0)
        K2c = vc.getTraitCovar(1)
    else:
        assert K1c.shape[0] == P, 'K1c: dimensions dismatch'
        assert K1c.shape[1] == P, 'K1c: dimensions dismatch'
        assert K2c.shape[0] == P, 'K2c: dimensions dismatch'
        assert K2c.shape[1] == P, 'K2c: dimensions dismatch'

    # 2. run kroneckerLMM for null model
    lmm = limix_legacy.deprecated.CKroneckerLMM()
    lmm.setK1r(K1r)
    lmm.setK1c(K1c)
    lmm.setK2r(K2r)
    lmm.setK2c(K2c)
    lmm.setSNPs(snps)
    # add covariates
    for ic in range(len(Acovs)):
        lmm.addCovariates(covs[ic], Acovs[ic])
    lmm.setPheno(phenos)

    # delta serch on alt. model?
    if searchDelta:
        lmm.setNumIntervalsAlt(NumIntervalsDeltaAlt)
        lmm.setNumIntervals0_inter(NumIntervalsDeltaAlt)
    else:
        lmm.setNumIntervalsAlt(0)
        lmm.setNumIntervals0_inter(0)

    lmm.setNumIntervals0(NumIntervalsDelta0)
    # add SNP design
    lmm.setSNPcoldesign0_inter(Asnps0[0])
    for iA in range(len(Asnps1)):
        lmm.setSNPcoldesign(Asnps1[iA])
        lmm.process()

        pvAlt[iA, :] = lmm.getPv()[0]
        pv[iA, :] = lmm.getPv()[1]
        pv0 = lmm.getPv()[2][np.newaxis, :]
    if return_lmm:
        return pv, pv0, pvAlt, lmm
    else:
        return pv, pv0, pvAlt


def qtl_test_interaction_lmm(snps,
                             pheno,
                             Inter,
                             Inter0=None,
                             covs=None,
                             K=None,
                             test='lrt'):
    r"""
    Wrapper function for single-variant interaction test using LMMs.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        Inter (ndarray, optional):
            (`N`, `I`) ndarray of `I` interaction variables to be tested
            for `N` individuals.
        Inter0 (ndarray, optional):
            (`N`, `I0`) ndarray of `I0` interaction variables to be included
            for `N` individuals to be included both in the alternative and
            in the null model.
            By default `Inter0` is a (`N`, `1`) array of ones.
        covs (ndarray, optional):
            (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        K (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients (optional)
            If not provided, then standard linear regression is considered.
        test ({'lrt', 'f'}, optional):
            test statistic.
            'lrt' for likelihood ratio test (default) or 'f' for F-test.

    Returns:
        :class:`limix_legacy.deprecated.CInteractLMM()`:
            limix CInteractLMM object

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import dot, eye, ones, concatenate
            >>> from limix.qtl import qtl_test_interaction_lmm
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, 1)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>> kinship+= 1e-4 * eye(N)
            >>>
            >>> #environments to include in null and alt models
            >>> Inter0 = ones((N,1))
            >>>
            >>> #environments to include in alt model
            >>> Inter = random.randn(N,2)
            >>>
            >>> #environments are also included as fixed effect covariates
            >>> mean = ones([N, 1])
            >>> covs = concatenate([mean, Inter], 1)
            >>>
            >>> #interaction test
            >>> lmi = qtl_test_interaction_lmm(snps, pheno, Inter=Inter,
            ...                                Inter0=Inter0)
            >>> pvi = lmi.getPv()
            >>>
            >>> print(pvi.shape)
            (1, 1000)
            >>> print(pvi[:,:4]) # doctest: +SKIP
            [[0.8179 0.2185 0.3338 0.077 ]]
    """

    import limix_legacy.deprecated
    import limix_legacy.deprecated as dlimix_legacy
    N = snps.shape[0]
    if covs is None:
        covs = np.ones((N, 1))
    if K is None:
        K = np.eye(N)
    if Inter0 is None:
        Inter0 = np.ones([N, 1])
    ok = pheno.shape[0] == N and K.shape[0] == N
    ok = ok and K.shape[1] == N and covs.shape[0] == N
    ok = ok and Inter0.shape[0] == N and Inter.shape[0] == N
    if not ok:
        raise ValueError("shapes missmatch")
    lmi = limix_legacy.deprecated.CInteractLMM()
    lmi.setK(K)
    lmi.setSNPs(snps)
    lmi.setPheno(pheno)
    lmi.setCovs(covs)
    lmi.setInter0(Inter0)
    lmi.setInter(Inter)
    if test == 'lrt':
        lmi.setTestStatistics(lmi.TEST_LRT)
    elif test == 'f':
        lmi.setTestStatistics(lmi.TEST_F)
    else:
        print(test)
        raise NotImplementedError("only f or lrt are implemented")
    lmi.process()
    return lmi


""" MULTI LOCUS MODEL """


def forward_lmm(snps,
                pheno,
                K=None,
                covs=None,
                qvalues=False,
                threshold=5e-8,
                maxiter=2,
                test='lrt',
                verbose=None,
                **kw_args):
    r"""
    Wrapper function for univariate single-variant test with forward selection

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        K (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients (optional)
            If not provided, then standard linear regression is considered.
        covs (ndarray, optional):
            (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        qvalues (bool, optional):
            if True, ``threshold`` is set to Storey qvalues to control FDR.
            (default is False)
        threshold (float, optional):
            P-value thrashold for inclusion in forward selection (default 5e-8).
            If ``qvalues=True``, the threshold is set to to Storey qvalues.
        maxiter (int, optional):
            maximum number of interaction scans. First scan is
            without inclusion, so maxiter-1 inclusions can be performed.
            (default 2)
        test ({'lrt', 'f'}, optional):
            test statistic.
            'lrt' for likelihood ratio test (default) or 'f' for F-test.
        verbose (bool, optional):
            print verbose output. (default is False)

    Returns:
        (tuple): tuple containing:
            - **lmm** (*:class:`limix.qtl.LMM`*): LIMIX LMM object
            - **res** (*dict*): {**iadded**, **pvadded**, **pvall**}.
              **iadded** is an ndarray of indices of SNPs included in order
              of inclusion, **pvadded** is an ndarray of Pvalues obtained by
              the included SNPs in iteration and **pvall** is a  (`Nadded`, `S`)
              ndarray of Pvalues for all iterations

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from numpy import dot, eye, ones
            >>> from limix.qtl import forward_lmm
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, 1)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>> kinship+= 1e-4 * eye(N)
            >>>
            >>> #forward lmm
            >>> lmm, res = forward_lmm(snps, pheno, K=kinship, threshold=1.)
            >>>
            >>> print(res['pvall'].shape)
            (2, 1000)
            >>> print(res['pvall'][:,:4]) # doctest: +SKIP
            [[0.8571 0.4668 0.5872 0.5589]
             [0.77   0.4226 0.6165 0.8727]]
    """

    import limix_legacy.deprecated
    import limix_legacy.deprecated as dlimix_legacy
    verbose = dlimix_legacy.getVerbose(verbose)

    if K is None:
        K = np.eye(snps.shape[0])
    if covs is None:
        covs = np.ones((snps.shape[0], 1))
    # assert single trait
    assert pheno.shape[1] == 1, 'forward_lmm only supports single phenotypes'

    lm = qtl_test_lmm(snps, pheno, K=K, covs=covs, test=test, **kw_args)
    pvall = []
    pv = lm.getPv().ravel()
    # hack to avoid issues with degenerate pv
    pv[sp.isnan(pv)] = 1
    pvall.append(pv)
    imin = pv.argmin()
    niter = 1
    # start stuff
    iadded = []
    pvadded = []
    qvadded = []
    if qvalues:
        assert pv.shape[
            0] == 1, "This is untested with the fdr package. pv.shape[0]==1 failed"
        qvall = []
        qv = qvalues(pv)
        qvall.append(qv)
        score = qv.min()
    else:
        score = pv.min()
    while (score < threshold) and niter < maxiter:
        t0 = time.time()
        iadded.append(imin)
        pvadded.append(pv[imin])
        if qvalues:
            qvadded.append(qv[0, imin])
        covs = np.concatenate((covs, snps[:, imin:(imin + 1)]), 1)
        lm.setCovs(covs)
        lm.process()
        pv = lm.getPv().ravel()
        pv[sp.isnan(pv)] = 1
        pvall.append(pv)
        imin = pv.argmin()
        if qvalues:
            qv = qvalues(pv)
            qvall[niter:niter + 1, :] = qv
            score = qv.min()
        else:
            score = pv.min()
        t1 = time.time()
        if verbose:
            print(("finished GWAS testing in %.2f seconds" % (t1 - t0)))
        niter = niter + 1
    RV = {}
    RV['iadded'] = iadded
    RV['pvadded'] = pvadded
    RV['pvall'] = np.array(pvall)
    if qvalues:
        RV['qvall'] = np.array(qvall)
        RV['qvadded'] = qvadded
    return lm, RV


# TOOD: use **kw_args to forward params.. see below
def forward_lmm_kronecker(snps,
                          phenos,
                          Asnps=None,
                          Acond=None,
                          K1r=None,
                          K1c=None,
                          K2r=None,
                          K2c=None,
                          covs=None,
                          Acovs=None,
                          threshold=5e-8,
                          maxiter=2,
                          qvalues=False,
                          update_covariances=False,
                          verbose=None,
                          **kw_args):
    """
    Kronecker fixed effects test with forward selection

    Args:
        snps:   [N x S] np.array of S SNPs for N individuals (test SNPs)
        pheno:  [N x P] np.array of 1 phenotype for N individuals
        K:      [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        covs:   [N x D] np.array of D covariates for N individuals
        threshold:      (float) P-value thrashold for inclusion in forward selection (default 5e-8)
        maxiter:        (int) maximum number of interaction scans. First scan is
                        without inclusion, so maxiter-1 inclusions can be performed. (default 2)
        qvalues:        Use q-value threshold and return q-values in addition (default False)
        update_covar:   Boolean indicator if covariances should be re-estimated after each forward step (default False)

    Returns:
        lm:             lmix LMMi object
        resultStruct with elements:
            iadded:         array of indices of SNPs included in order of inclusion
            pvadded:        array of Pvalues obtained by the included SNPs in iteration
                            before inclusion
            pvall:     [Nadded x S] np.array of Pvalues for all iterations.
        Optional:      corresponding q-values
            qvadded
            qvall
    """
    import limix_legacy.deprecated
    import limix_legacy.deprecated as dlimix_legacy
    verbose = dlimix_legacy.getVerbose(verbose)
    # 0. checks
    N = phenos.shape[0]
    P = phenos.shape[1]

    if K1r is None:
        K1r = np.dot(snps, snps.T)
    else:
        assert K1r.shape[0] == N, 'K1r: dimensions dismatch'
        assert K1r.shape[1] == N, 'K1r: dimensions dismatch'

    if K2r is None:
        K2r = np.eye(N)
    else:
        assert K2r.shape[0] == N, 'K2r: dimensions dismatch'
        assert K2r.shape[1] == N, 'K2r: dimensions dismatch'

    covs, Acovs = _updateKronCovs(covs, Acovs, N, P)

    if Asnps is None:
        Asnps = [np.ones([1, P])]
    if (not isinstance(Asnps, list)):
        Asnps = [Asnps]
    assert len(Asnps) > 0, "need at least one Snp design matrix"

    if Acond is None:
        Acond = Asnps
    if (not isinstance(Acond, list)):
        Acond = [Acond]
    assert len(Acond) > 0, "need at least one Snp design matrix"

    # 1. run GP model to infer suitable covariance structure
    if K1c is None or K2c is None:
        vc = _estimateKronCovariances(
            phenos=phenos,
            K1r=K1r,
            K2r=K2r,
            K1c=K1c,
            K2c=K2c,
            covs=covs,
            Acovs=Acovs,
            **kw_args)
        K1c = vc.getTraitCovar(0)
        K2c = vc.getTraitCovar(1)
    else:
        vc = None
        assert K1c.shape[0] == P, 'K1c: dimensions dismatch'
        assert K1c.shape[1] == P, 'K1c: dimensions dismatch'
        assert K2c.shape[0] == P, 'K2c: dimensions dismatch'
        assert K2c.shape[1] == P, 'K2c: dimensions dismatch'
    t0 = time.time()
    lm, pv = qtl_test_lmm_kronecker(
        snps=snps,
        phenos=phenos,
        Asnps=Asnps,
        K1r=K1r,
        K2r=K2r,
        K1c=K1c,
        K2c=K2c,
        covs=covs,
        Acovs=Acovs)

    # get pv
    # start stuff
    iadded = []
    pvadded = []
    qvadded = []
    time_el = []
    pvall = []
    qvall = None
    t1 = time.time()
    if verbose:
        print(("finished GWAS testing in %.2f seconds" % (t1 - t0)))
    time_el.append(t1 - t0)
    pvall.append(pv)
    imin = np.unravel_index(pv.argmin(), pv.shape)
    score = pv[imin].min()
    niter = 1
    if qvalues:
        assert pv.shape[
            0] == 1, "This is untested with the fdr package. pv.shape[0]==1 failed"
        qvall = []
        qv = qvalues(pv)
        qvall.append(qv)
        score = qv[imin]
    # loop:
    while (score < threshold) and niter < maxiter:
        t0 = time.time()
        pvadded.append(pv[imin])
        iadded.append(imin)
        if qvalues:
            qvadded.append(qv[imin])
        if update_covariances and vc is not None:
            vc.addFixedTerm(snps[:, imin[1]:(imin[1] + 1)], Acond[imin[0]])
            vc.setScales(
            )  # CL: don't know what this does, but findLocalOptima crashes becahuse vc.noisPos=None
            vc.findLocalOptima(fast=True)
            K1c = vc.getTraitCovar(0)
            K2c = vc.getTraitCovar(1)
            lm.setK1c(K1c)
            lm.setK2c(K2c)
        lm.addCovariates(snps[:, imin[1]:(imin[1] + 1)], Acond[imin[0]])
        for i in range(len(Asnps)):
            # add SNP design
            lm.setSNPcoldesign(Asnps[i])
            lm.process()
            pv[i, :] = lm.getPv()[0]
        pvall.append(pv.ravel())
        imin = np.unravel_index(pv.argmin(), pv.shape)
        if qvalues:
            qv = qvalues(pv)
            qvall[niter:niter + 1, :] = qv
            score = qv[imin].min()
        else:
            score = pv[imin].min()
        t1 = time.time()
        if verbose:
            print(("finished GWAS testing in %.2f seconds" % (t1 - t0)))
        time_el.append(t1 - t0)
        niter = niter + 1
    RV = {}
    RV['iadded'] = iadded
    RV['pvadded'] = pvadded
    RV['pvall'] = np.array(pvall)
    RV['time_el'] = time_el
    if qvalues:
        RV['qvall'] = qvall
        RV['qvadded'] = qvadded
    return lm, RV


""" INTERNAL """


def _estimateKronCovariances(phenos,
                             K1r=None,
                             K1c=None,
                             K2r=None,
                             K2c=None,
                             covs=None,
                             Acovs=None,
                             trait_covar_type='freeform',
                             rank=1,
                             lambd=None,
                             verbose=True,
                             init_method='random',
                             old_opt=True):
    """
    estimates the background covariance model before testing

    Args:
        phenos: [N x P] np.array of P phenotypes for N individuals
        K1r:    [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        K1c:    [P x P] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        K2r:    [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        K2c:    [P x P] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        covs:           list of np.arrays holding covariates. Each covs[i] has one corresponding Acovs[i]
        Acovs:          list of np.arrays holding the phenotype design matrices for covariates.
                        Each covs[i] has one corresponding Acovs[i].
        trait_covar_type:     type of covaraince to use. Default 'freeform'. possible values are
                        'freeform': free form optimization,
                        'fixed': use a fixed matrix specified in covar_K0,
                        'diag': optimize a diagonal matrix,
                        'lowrank': optimize a low rank matrix. The rank of the lowrank part is specified in the variable rank,
                        'lowrank_id': optimize a low rank matrix plus the weight of a constant diagonal matrix. The rank of the lowrank part is specified in the variable rank,
                        'lowrank_diag': optimize a low rank matrix plus a free diagonal matrix. The rank of the lowrank part is specified in the variable rank,
                        'block': optimize the weight of a constant P x P block matrix of ones,
                        'block_id': optimize the weight of a constant P x P block matrix of ones plus the weight of a constant diagonal matrix,
                        'block_diag': optimize the weight of a constant P x P block matrix of ones plus a free diagonal matrix,
        rank:           rank of a possible lowrank component (default 1)

    Returns:
        VarianceDecomposition object
    """
    import limix_legacy.modules.varianceDecomposition as VAR

    # from . import varianceDecomposition as VAR
    if verbose:
        print(".. Estimating trait covariances")
    vc = VAR.VarianceDecomposition(phenos)
    if K1r is not None:
        vc.addRandomEffect(K1r, trait_covar_type='freeform', rank=rank)
    if K2r is not None:
        # TODO: fix this; forces second term to be the noise covariance
        vc.addRandomEffect(
            is_noise=True, K=K2r, trait_covar_type='freeform', rank=rank)
    for ic in range(len(Acovs)):
        vc.addFixedEffect(covs[ic], Acovs[ic])
    start = time.time()
    if old_opt:
        conv = vc.optimize(fast=True)
    elif lambd is not None:
        conv = vc.optimize(
            init_method=init_method, verbose=verbose, lambd=lambd)
    else:
        conv = vc.optimize(init_method=init_method, verbose=verbose)
    assert conv, "Variance Decomposition has not converged"
    time_el = time.time() - start
    if verbose:
        print(("Done in %.2f s" % time_el))
    return vc


def _updateKronCovs(covs, Acovs, N, P):
    """
    make sure that covs and Acovs are lists
    """
    if (covs is None) and (Acovs is None):
        covs = [np.ones([N, 1])]
        Acovs = [np.eye(P)]

    if Acovs is None or covs is None:
        raise Exception("Either Acovs or covs is None, while the other isn't")

    if (not isinstance(Acovs, list)) and (not isinstance(covs, list)):
        Acovs = [Acovs]
        covs = [covs]
    if (not isinstance(covs, list)) or (not isinstance(
            Acovs, list)) or (len(covs) != len(Acovs)):
        raise Exception(
            "Either Acovs or covs is not a list or they missmatch in length")
    return covs, Acovs


# TODO: we need to fix. THis does not work as interact_GxE is not existing
# I vote we also use **kw_args to forward parameters to interact_Gxe?
def qtl_test_interaction_GxG(pheno,
                             snps1,
                             snps2=None,
                             K=None,
                             covs=None,
                             test='lrt'):
    """
    Epistasis test between two sets of SNPs

    Args:
        pheno:  [N x 1] np.array of 1 phenotype for N individuals
        snps1:  [N x S1] np.array of S1 SNPs for N individuals
        snps2:  [N x S2] np.array of S2 SNPs for N individuals
        K:      [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        covs:   [N x D] np.array of D covariates for N individuals
        test:    'lrt' for likelihood ratio test (default) or 'f' for F-test

    Returns:
        pv:     [S2 x S1] np.array of P values for epistasis tests beten all SNPs in
                snps1 and snps2
    """
    if K is None:
        K = np.eye(N)
    N = snps1.shape[0]
    if snps2 is None:
        snps2 = snps1
    return qtl_test_interaction_GxE_1dof(
        snps=snps1, pheno=pheno, env=snps2, covs=covs, K=K, test=test)


def qtl_test_interaction_GxE_1dof(snps,
                                  pheno,
                                  env,
                                  K=None,
                                  covs=None,
                                  test='lrt',
                                  verbose=None):
    """
    Univariate GxE fixed effects interaction linear mixed model test for all
    pairs of SNPs and environmental variables.

    Args:
        snps:   [N x S] np.array of S SNPs for N individuals
        pheno:  [N x 1] np.array of 1 phenotype for N individuals
        env:    [N x E] np.array of E environmental variables for N individuals
        K:      [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        covs:   [N x D] np.array of D covariates for N individuals
        test:    'lrt' for likelihood ratio test (default) or 'f' for F-test
        verbose: print verbose output? (False)

    Returns:
        pv:     [E x S] np.array of P values for interaction tests between all
                E environmental variables and all S SNPs
    """
    import limix_legacy.deprecated as dlimix_legacy
    verbose = dlimix_legacy.getVerbose(verbose)
    N = snps.shape[0]
    if K is None:
        K = np.eye(N)
    if covs is None:
        covs = np.ones((N, 1))
    ok = env.shape[0] == N and pheno.shape[0] == N
    ok = ok and K.shape[0] == N and K.shape[1] == N
    ok = ok and covs.shape[0] == N
    if not ok:
        raise ValueError("shapes missmatch")
    Inter0 = np.ones((N, 1))
    pv = np.zeros((env.shape[1], snps.shape[1]))
    if verbose:
        print(("starting %i interaction scans for %i SNPs each." %
               (env.shape[1], snps.shape[1])))
    t0 = time.time()
    for i in range(env.shape[1]):
        t0_i = time.time()
        cov_i = np.concatenate((covs, env[:, i:(i + 1)]), 1)
        lm_i = qtl_test_interaction_lmm(
            snps=snps,
            pheno=pheno,
            covs=cov_i,
            Inter=env[:, i:(i + 1)],
            Inter0=Inter0,
            test=test)
        pv[i, :] = lm_i.getPv()[0, :]
        t1_i = time.time()
        if verbose:
            print(("Finished %i out of %i interaction scans in %.2f seconds." %
                   ((i + 1), env.shape[1], (t1_i - t0_i))))
    t1 = time.time()
    print((
        "-----------------------------------------------------------\nFinished all %i interaction scans in %.2f seconds."
        % (env.shape[1], (t1 - t0))))
    return pv


def phenSpecificEffects(snps, pheno1, pheno2, K=None, covs=None, test='lrt'):
    """
    Univariate fixed effects interaction test for phenotype specific SNP effects

    Args:
        snps:   [N x S] np.array of S SNPs for N individuals (test SNPs)
        pheno1: [N x 1] np.array of 1 phenotype for N individuals
        pheno2: [N x 1] np.array of 1 phenotype for N individuals
        K:      [N x N] np.array of LMM-covariance/kinship koefficients (optional)
                        If not provided, then linear regression analysis is performed
        covs:   [N x D] np.array of D covariates for N individuals
        test:    'lrt' for likelihood ratio test (default) or 'f' for F-test

    Returns:
        limix LMM object
    """
    N = snps.shape[0]
    if K is None:
        K = np.eye(N)
    assert (pheno1.shape[1] == pheno2.shape[1]
            ), "Only consider equal number of phenotype dimensions"
    if covs is None:
        covs = np.ones(N, 1)
    ok = pheno1.shape[1] == 1 and pheno2.shape[1] == 1
    ok = ok and pheno1.shape[0] == N and pheno2.shape[0] == N
    ok = ok and K.shape[0] == N and K.shape[1] == N
    ok = ok and covs.shape[0] == N
    if not ok:
        raise ValueError("shapes missmatch")
    Inter = np.zeros((N * 2, 1))
    Inter[0:N, 0] = 1
    Inter0 = np.ones((N * 2, 1))
    Yinter = np.concatenate((pheno1, pheno2), 0)
    Xinter = np.tile(snps, (2, 1))
    Covitner = np.tile(covs(2, 1))
    lm = qtl_test_interaction_lmm(
        snps=Xinter,
        pheno=Yinter,
        covs=Covinter,
        Inter=Inter,
        Inter0=Inter0,
        test=test)
    return lm



import os
import sys
import csv
import numpy as np
from optparse import OptionParser
import time
import limix
from .read_utils import readNullModelFile
from .read_utils import readWindowsFile
from .read_utils import readCovarianceMatrixFile
from .read_utils import readCovariatesFile
from .read_utils import readPhenoFile
from limix.data import BedReader
from limix.util import unique_variants as f_uni_variants
import scipy as sp
import warnings
import pandas as pd


def scan(
        bfile,
        Y,
        cov,
        null,
        sets,
        i0,
        i1,
        perm_i,
        resfile,
        F,
        colCovarType_r='lowrank',
        rank_r=1,
        factr=1e7,
        unique_variants=False,
        standardize=False):

    if perm_i is not None:
        print(('Generating permutation (permutation %d)' % perm_i))
        np.random.seed(perm_i)
        perm = np.random.permutation(Y.shape[0])

    mtSet = limix.MTSet(
        Y=Y,
        S_R=cov['eval'],
        U_R=cov['evec'],
        F=F,
        rank=rank_r)
    mtSet.setNull(null)

    reader = BedReader(bfile)

    wnd_ids = sp.arange(i0, i1)
    LLR = sp.zeros(sets.shape[0])
    for wnd_i in wnd_ids:

        _set = sets.ix[wnd_i]
        print('.. set %d: %s' % (wnd_i, _set['setid']))

        Xr = reader.getGenotypes(pos_start=_set['start'],
                                 pos_end=_set['end'],
                                 chrom=_set['chrom'],
                                 impute=True)

        if unique_variants:
            Xr = f_uni_variants(Xr)

        if standardize:
            Xr -= Xr.mean(0)
            Xr /= Xr.std(0)
        else:
            # encoding minor as 0
            p = 0.5 * Xr.mean(0)
            Xr[:, p > 0.5] = 2 - Xr[:, p > 0.5]

        if perm_i is not None:
            Xr = Xr[perm, :]

        # multi trait set test fit
        RV = mtSet.optimize(Xr, factr=factr)
        LLR[wnd_i] = RV['LLR'][0]

    # export results
    sets['LLR'] = LLR
    sets.to_csv(resfile, sep='\t', index=False)


def analyze(options):

    # load data
    print('import data')
    if options.cfile is None:
        cov = {'eval': None, 'evec': None}
        warnings.warn(
            'warning: cfile not specifed, a one variance compoenent model will be considered')
    else:
        cov = readCovarianceMatrixFile(options.cfile, readCov=False)
    Y = readPhenoFile(options.pfile, idx=options.trait_idx)
    null = readNullModelFile(options.nfile)

    sets = pd.DataFrame.from_csv(options.wfile + '.wnd',
                                 sep='\t',
                                 index_col=None)

    F = None
    if options.ffile:
        F = readCovariatesFile(options.ffile)
        # null['params_mean'] = sp.loadtxt(options.nfile + '.f0')

    if F is not None:
        assert Y.shape[0] == F.shape[0], 'dimensions mismatch'

    if options.i0 is None:
        options.i0 = 1
    if options.i1 is None:
        options.i1 = sets.shape[0]

    # name of output file
    if options.perm_i is not None:
        res_dir = os.path.join(options.resdir, 'perm%d' % options.perm_i)
    else:
        res_dir = os.path.join(options.resdir, 'test')
    if not os.path.exists(res_dir):
        os.makedirs(res_dir)
    n_digits = len(str(sets.shape[0]))
    fname = str(options.i0).zfill(n_digits)
    fname += '_' + str(options.i1).zfill(n_digits) + '.res'
    resfile = os.path.join(res_dir, fname)

    # analysis
    t0 = time.time()
    scan(
        options.bfile,
        Y,
        cov,
        null,
        sets,
        options.i0,
        options.i1,
        options.perm_i,
        resfile,
        F,
        options.colCovarType_r,
        options.rank_r,
        options.factr,
        options.unique_variants,
        options.standardize)
    t1 = time.time()
    print(('... finished in %s seconds' % (t1 - t0)))



def see_image(filepath):
    import matplotlib.pyplot as plt

    plt.imshow(plt.imread(filepath))
    axes = plt.gca()
    plt.tight_layout()
    axes.set_position([0, 0, 1, 1])
    axes.xaxis.set_visible(False)
    axes.yaxis.set_visible(False)
    plt.show()



import pandas as pd
import scipy as sp

from .mvset import MvSetTest
from .mvsetfull import MvSetTestFull
from .mvsetinc import MvSetTestInc


def fit_iSet(Y=None,
             Xr=None,
             F=None,
             Rg=None,
             Ug=None,
             Sg=None,
             Ie=None,
             n_nulls=10,
             factr=1e7):
    """
    Fit interaction set test (iSet).

    Args:
        Y (ndarray):
            For complete design, the phenotype ndarray `Y`
            for `N` samples and `C` contexts has shape (`N`, `C`).
            For stratified design, the phenotype ndarray `Y` has
            shape (`N`, `1`) (each individual is phenotyped in only one context -
            ``Ie`` specifies in which context each individuals has been phenotyped).
        Xr (ndarray):
            (`N`, `S`) genotype values for `N` samples and `S` variants
            (defines the set component)
        F (ndarray, optional):
            (`N`, `K`) ndarray of `K` covariates for `N` individuals.
            By default, ``F`` is a (`N`, `1`) array of ones.
        Rg (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients.
            ``Ug`` and ``Sg`` can be provided instead of ``Rg``.
            If neither ``Rg`` nor ``Ug`` and ``Sg`` are provided,
            the null models has iid normal residuals.
        Ug (ndarray, optional):
            (`N`, `N`) ndarray of eigenvectors of ``Rg``.
            ``Ug`` and ``Sg`` can be provided instead of ``Rg``.
            If neither ``Rg`` nor ``Ug`` and ``Sg`` are provided,
            iid normal residuals are considered.
        Sg (ndarray, optional):
            (`N`, ) ndarray of eigenvalues of ``Rg``.
            ``Ug`` and ``Sg`` can be provided instead of ``Rg``.
            If neither ``Rg`` nor ``Ug`` and ``Sg`` are provided,
            iid normal residuals are considered.
        Ie (ndarry, optional):
            (`N`, `1`) boolean ndarray indicator for analysis of
            stratified designs.
            More specifically ``Ie`` specifies in which context each
            individuals has been phenotyped.
            Needs to be specified for analysis of stratified designs.
        n_nulls (ndarray, optional):
            number of parametric bootstrap. This parameter determines
            the minimum P value that can be estimated.
            The default value is 10.
        factr (float, optional):
            optimization paramenter that determines the accuracy of the solution
            (see scipy.optimize.fmin_l_bfgs_b for more details).

    Returns:
        (tuple): tuple containing:
            - **df** (*:class:`pandas.DataFrame`*):
              contains test statistcs of mtSet, iSet, and iSet-GxC tests and
              the variance exaplained by persistent, GxC and
              heterogeneity-GxC effects.
            - **df0** (*:class:`pandas.DataFrame`*):
              contains null test statistcs of mtSet, iSet, and iSet-GxC tests.

    Examples
    --------

        This example shows how to fit iSet when considering complete designs and
        modelling population structure/relatedness by introducing the top principle
        components of the genetic relatedness matrix (``pc_rrm``) as fixed effects.

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.iset import fit_iSet
            >>> from numpy import ones, concatenate
            >>> import scipy as sp
            >>>
            >>> random = RandomState(1)
            >>> sp.random.seed(0)
            >>>
            >>> N = 100
            >>> C = 2
            >>> S = 4
            >>>
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, C)
            >>> mean = ones((N, 1))
            >>> pc_rrm = random.randn(N, 4)
            >>> covs = concatenate([mean, pc_rrm], 1)
            >>>
            >>> df, df0 = fit_iSet(Y=pheno, Xr=snps, F=covs, n_nulls=2)
            >>>
            >>> print(df.round(3).T)
                                       0
            Heterogeneity-GxC var  0.000
            Persistent Var         0.005
            Rescaling-GxC Var      0.005
            iSet LLR               0.137
            iSet-het LLR          -0.000
            mtSet LLR              0.166

        This example shows how to fit iSet when considering complete designs
        and modelling population structure/relatedness using the full
        genetic relatedness matrix.

        .. doctest::

            >>> from numpy import dot, eye
            >>> random = RandomState(1)
            >>> sp.random.seed(0)
            >>>
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>> kinship+= 1e-4 * eye(N)
            >>>
            >>> df, df0 = fit_iSet(Y=pheno, Xr=snps, Rg=kinship, n_nulls=2)
            >>>
            >>> print(df.round(3).T)
                                       0
            Heterogeneity-GxC var  0.000
            Persistent Var         0.005
            Rescaling-GxC Var      0.005
            iSet LLR               1.098
            iSet-het LLR           1.014
            mtSet LLR              0.154

        This example shows how to fit iSet when considering stratified designs
        and modelling population structure/relatedness by introducing
        the top principle components of the genetic relatedness matrix
        (``pc_rrm``) as fixed effects.
        iSet does not support models with full genetic relatedness matrix
        for stratified designs.

        .. doctest::

            >>> random = RandomState(1)
            >>> sp.random.seed(0)
            >>>
            >>> pheno = random.randn(N, 1)
            >>> Ie = random.randn(N)<0.
            >>>
            >>> df, df0 = fit_iSet(Y=pheno, Xr=snps, F=covs, Ie=Ie, n_nulls=2)
            >>>
            >>> print(df.round(3).T)
                                       0
            Heterogeneity-GxC var -0.000
            Persistent Var         0.064
            Rescaling-GxC Var      0.006
            iSet LLR               0.648
            iSet-het LLR           0.000
            mtSet LLR              1.177

        For more info and examples see the `iSet tutorial`_.

        .. _iSet tutorial: https://github.com/limix/limix-tutorials/tree/master/iSet

    """
    # data
    noneNone = Sg is not None and Ug is not None
    bgRE = Rg is not None or noneNone
    # fixed effect
    msg = 'The current implementation of the full rank iSet'
    msg += ' does not support covariates.'
    msg += ' We reccommend to regress out covariates and'
    msg += ' subsequently quantile normalize the phenotypes'
    msg += ' to a normal distribution prior to use mtSet/iSet.'
    msg += ' This can be done within the LIMIX framework using'
    msg += ' the methods limix.util.preprocess.regressOut and'
    msg += ' limix.util.preprocess.gaussianize'
    assert not (F is not None and bgRE), msg
    # strat
    strat = Ie is not None
    msg = 'iSet for interaction analysis of stratified populations '
    msg += 'using contextual variables does not support random effect '
    msg += 'correction for confounding. '
    msg += 'Please use the fixed effects to correct for confounding. '
    assert not (strat and bgRE), msg

    # check inputs for strat design
    if strat:
        assert Y.shape[1] == 1, 'Y has not the right shape'
        valid_env_ids = (sp.unique(Ie) == sp.array([0, 1])).all()
        assert valid_env_ids, 'The provided Ie is not valid'

    # phenotype is centered for 2 random effect model
    if bgRE:
        Yc = Y - Y.mean(0)

    # define mtSet
    if bgRE:
        mvset = MvSetTestFull(Y=Yc, Xr=Xr, Rg=Rg, Ug=Ug, Sg=Sg, factr=factr)
    elif strat:
        mvset = MvSetTestInc(Y=Y, Xr=Xr, F=F, Ie=Ie, factr=factr)
    else:
        mvset = MvSetTest(Y=Y, Xr=Xr, F=F, factr=factr)

    RV = {}
    RV['mtSet LLR'] = mvset.assoc()
    RV['iSet LLR'] = mvset.gxe()
    RV['iSet-het LLR'] = mvset.gxehet()
    RV['Persistent Var'] = mvset.info['full']['var_r'][0]
    RV['Rescaling-GxC Var'] = mvset.info['full']['var_r'][1]
    RV['Heterogeneity-GxC var'] = mvset.info['full']['var_r'][2]
    df = pd.DataFrame(RV)

    RV0 = {}
    RV0['mtSet LLR0'] = mvset.assoc_null(n_nulls=n_nulls)
    RV0['iSet LLR0'] = mvset.gxe_null(n_nulls=n_nulls)
    RV0['iSet-het LLR0'] = mvset.gxehet_null(n_nulls=n_nulls)
    df0 = pd.DataFrame(RV0)

    return df, df0



from numpy import dot, exp, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose
from numpy_sugar.linalg import economic_qs

from limix.heritability import estimate


def test_heritability_estimate_binomial():
    random = RandomState(0)
    nsamples = 50

    X = random.randn(50, 2)
    G = random.randn(50, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for j in range(ntrials[i]):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    y = (successes, ntrials)

    assert_allclose(
        estimate(y, 'binomial', K, verbose=False), 0.8937449693087052)


def test_heritability_estimate_poisson():
    random = RandomState(0)
    nsamples = 50

    G = random.randn(50, 100)
    K = dot(G, G.T)
    z = dot(G, random.randn(100)) / sqrt(100)
    y = random.poisson(exp(z))

    assert_allclose(
        estimate(y, 'poisson', K, verbose=False), 0.6998737749318877)



r"""
*************
Dataset utils
*************

- :class:`.BedReader`
- :class:`.query_and`
- :class:`.build_geno_query`

Public interface
^^^^^^^^^^^^^^^^
"""

from .bed_reader import BedReader
from .query_utils import query_and
from .query_utils import build_geno_query
from .geno_iterator import GIter

__all__ = ['BedReader', 'query_and', 'build_geno_query']



import cmd
import shlex
from argparse import ArgumentParser

from . import limix as limix_cmd


class ILimix(cmd.Cmd):
    def do_see(self, cmdline):
        p = limix_cmd.see_parser(ArgumentParser())
        args = p.parse_args(shlex.split(cmdline))

        limix_cmd.do_see(args)

    def do_EOF(self, _):
        return True

    def do_exit(self, *_):
        return True


def entry_point():
    ILimix().cmdloop()


if __name__ == '__main__':
    entry_point()



from limix.util import sets_from_bim
from limix.util import annotate_sets
from limix.io import read_plink
import time


def definesets(opt):

    assert opt.bfile is not None, 'Please specify a bfile.'

    """ precomputing the windows """
    if opt.wfile is None:
        opt.wfile = os.path.split(
            opt.bfile)[-1] + '.%d' % opt.window_size
        warnings.warn('wfile not specifed, set to %s' % opt.wfile)

    t0 = time.time()

    (bim, fam, G) = read_plink(opt.bfile, verbose=False)

    if opt.sliding_wind:
        print('Precomputing windows')
        sets = sets_from_bim(bim,
                             size=opt.window_size,
                             step=opt.step,
                             chrom=opt.chrom,
                             minSnps=opt.minSnps,
                             maxSnps=opt.maxSnps)

    elif opt.filter_sets:
        assert opt.iwfile is not None, 'Please specify a iwfile.'
        print('Filtering windows')
        pdb.set_trace()
        sets = annotate_sets(sets0,
                             bim,
                             minSnps=opt.minSnps,
                             maxSnps=opt.maxSnps)

    sets.to_csv(opt.wfile + '.wnd', sep='\t', index=False)

    print(('Number of variants:', G.shape[0]))
    print(('Number of windows:', sets.shape[0]))
    t1 = time.time()
    print(('.. finished in %s seconds' % (t1 - t0)))



r"""
**********
I/O module
**********

PLINK reader
^^^^^^^^^^^^

.. autofunction:: limix.io.read_plink

HDF5 reader
^^^^^^^^^^^

.. :autoclass:: limix.io.h5data_fetcher

CSV reader
^^^^^^^^^^

.. autofunction:: limix.io.read_csv

GEN reader
^^^^^^^^^^

.. autofunction:: limix.io.read_gen

"""

from . import examples
from ._csv import read_csv
from .gen import read_gen
from . import hdf5
from .hdf5 import h5data_fetcher
from .plink import read_plink
from .util import file_type
from . import _csv as csv
from . import plink
from . import npy


def genotype_reader(*args, **kwargs):
    from limix_legacy.io import genotype_reader
    return genotype_reader(*args, **kwargs)


def phenotype_reader(*args, **kwargs):
    from limix_legacy.io import phenotype_reader
    return phenotype_reader(*args, **kwargs)


def data(*args, **kwargs):
    from limix_legacy.io import data
    return data(*args, **kwargs)


__all__ = [
    'read_plink', 'h5data_fetcher', 'read_csv', 'read_gen', 'examples',
    'genotype_reader', 'phenotype_reader', 'data', 'file_type', 'hdf5',
    'csv', 'plink', 'npy'
]



import re
import os
import subprocess
import pdb
import numpy as np
import numpy.linalg as la
from optparse import OptionParser
import time
import limix
from .read_utils import readBimFile
from .read_utils import readCovarianceMatrixFile
from .read_utils import readPhenoFile
from .read_utils import readCovariatesFile
from . import plink_reader
import scipy as sp
import warnings
import scipy.sparse.linalg as ssl


def computeCovarianceMatrixPlink(
        plink_path,
        out_dir,
        bfile,
        cfile,
        sim_type='RRM'):
    """
    computing the covariance matrix via plink
    """

    print("Using plink to create covariance matrix")
    cmd = '%s --bfile %s ' % (plink_path, bfile)

    if sim_type == 'RRM':
        # using variance standardization
        cmd += '--make-rel square '
    else:
        raise Exception('sim_type %s is not known' % sim_type)

    cmd += '--out %s' % (os.path.join(out_dir, 'plink'))

    subprocess.call(cmd, shell=True)

    # move file to specified file
    if sim_type == 'RRM':
        old_fn = os.path.join(out_dir, 'plink.rel')
        os.rename(old_fn, cfile + '.cov')

        old_fn = os.path.join(out_dir, 'plink.rel.id')
        os.rename(old_fn, cfile + '.cov.id')

    if sim_type == 'IBS':
        old_fn = os.path.join(out_dir, 'plink.mibs')
        os.rename(old_fn, cfile + '.cov')

        old_fn = os.path.join(out_dir, 'plink.mibs.id')
        os.rename(old_fn, cfile + '.cov.id')


def computePCsPlink(plink_path, k, out_dir, bfile, ffile):
    """
    computing the covariance matrix via plink
    """
    print("Using plink to compute principal components")
    cmd = '%s --bfile %s --pca %d ' % (plink_path, bfile, k)
    cmd += '--out %s' % (os.path.join(out_dir, 'plink'))
    subprocess.call(cmd, shell=True)
    plink_fn = os.path.join(out_dir, 'plink.eigenvec')
    M = sp.loadtxt(plink_fn, dtype=str)
    U = sp.array(M[:, 2:], dtype=float)
    U -= U.mean(0)
    U /= U.std(0)
    sp.savetxt(ffile, U)


def computePCsPython(out_dir, k, bfile, ffile):
    """ reading in """
    RV = plink_reader.readBED(bfile, useMAFencoding=True)
    X = np.ascontiguousarray(RV['snps'])

    """ normalizing markers """
    print('Normalizing SNPs...')
    p_ref = X.mean(axis=0) / 2.
    X -= 2 * p_ref

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X /= sp.sqrt(2 * p_ref * (1 - p_ref))

    hasNan = sp.any(sp.isnan(X), axis=0)
    if sp.any(hasNan):
        print(('%d SNPs have a nan entry. Exluding them for computing the covariance matrix.' % hasNan.sum()))
        X = X[:, ~hasNan]

    """ computing prinicipal components """
    U, S, Vt = ssl.svds(X, k=k)
    U -= U.mean(0)
    U /= U.std(0)
    U = U[:, ::-1]

    """ saving to output """
    np.savetxt(ffile, U, delimiter='\t', fmt='%.6f')


def computeCovarianceMatrixPython(out_dir, bfile, cfile, sim_type='RRM'):
    print("Using python to create covariance matrix. This might be slow. We recommend using plink instead.")

    if sim_type is not 'RRM':
        raise Exception('sim_type %s is not known' % sim_type)

    """ loading data """
    data = plink_reader.readBED(bfile, useMAFencoding=True)
    iid = data['iid']
    X = np.ascontiguousarray(data['snps'])
    N = X.shape[1]
    print(('%d variants loaded.' % N))
    print(('%d people loaded.' % X.shape[0]))
    """ normalizing markers """
    print('Normalizing SNPs...')
    p_ref = X.mean(axis=0) / 2.
    X -= 2 * p_ref

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        X /= sp.sqrt(2 * p_ref * (1 - p_ref))

    hasNan = sp.any(sp.isnan(X), axis=0)
    if sp.any(hasNan):
        print(('%d SNPs have a nan entry. Exluding them for computing the covariance matrix.' % hasNan.sum()))

    """ computing covariance matrix """
    print('Computing relationship matrix...')
    K = sp.dot(X[:, ~hasNan], X[:, ~hasNan].T)
    K /= 1. * N
    print('Relationship matrix calculation complete')
    print(('Relationship matrix written to %s.cov.' % cfile))
    print(('IDs written to %s.cov.id.' % cfile))

    """ saving to output """
    np.savetxt(cfile + '.cov', K, delimiter='\t', fmt='%.6f')
    np.savetxt(cfile + '.cov.id', iid, delimiter=' ', fmt='%s')


def computePCs(plink_path, k, bfile, ffile):
    """
    compute the first k principal components

    Input:
    k            :   number of principal components
    plink_path   :   plink path
    bfile        :   binary bed file (bfile.bed, bfile.bim and bfile.fam are required)
    ffile        :   name of output file
    """
    try:
        output = subprocess.check_output(
            '%s --version --noweb' %
            plink_path, shell=True)
        use_plink = float(output.split(' ')[1][1:-3]) >= 1.9
    except BaseException:
        use_plink = False

    assert bfile is not None, 'Path to bed-file is missing.'
    assert os.path.exists(bfile + '.bed'), '%s.bed is missing.' % bfile
    assert os.path.exists(bfile + '.bim'), '%s.bim is missing.' % bfile
    assert os.path.exists(bfile + '.fam'), '%s.fam is missing.' % bfile

    # create dir if it does not exist
    out_dir = os.path.split(ffile)[0]
    if out_dir != '' and (not os.path.exists(out_dir)):
        os.makedirs(out_dir)

    if use_plink:
        computePCsPlink(plink_path, k, out_dir, bfile, ffile)
    else:
        computePCsPython(out_dir, k, bfile, ffile)


def eighCovarianceMatrix(cfile):

    pdb.set_trace()
    S = np.loadtxt(cfile + '.cov.eval')  # S
    U = np.loadtxt(cfile + '.cov.evec')  # U
    pcs = U[:, :k]

    pcs -= pcs.mean(axis=0)
    pcs /= pcs.std(axis=0)

    np.savetxt(ffile, pcs, fmt='%.6f')


def computeCovarianceMatrix(plink_path, bfile, cfile, sim_type='RRM'):
    """
    compute similarity matrix using plink

    Input:
    plink_path   :   plink path
    bfile        :   binary bed file (bfile.bed, bfile.bim and bfile.fam are required)
    cfile        :   the covariance matrix will be written to cfile.cov and the corresponding identifiers
                         to cfile.cov.id. If not specified, the covariance matrix will be written to cfile.cov and
                         the individuals to cfile.cov.id in the current folder.
    sim_type     :   {IBS/RRM} are supported
    """
    try:
        output = subprocess.check_output(
            '%s --version --noweb' %
            plink_path, shell=True)
        m = re.match(b"^PLINK v(\d+\.\d+).*$", output)
        if m:
            use_plink = float(m.group(1)) >= 1.9
        else:
            use_plink = False
    except BaseException:
        use_plink = False

    assert bfile is not None, 'Path to bed-file is missing.'
    assert os.path.exists(bfile + '.bed'), '%s.bed is missing.' % bfile
    assert os.path.exists(bfile + '.bim'), '%s.bim is missing.' % bfile
    assert os.path.exists(bfile + '.fam'), '%s.fam is missing.' % bfile

    # create dir if it does not exist
    out_dir = os.path.split(cfile)[0]
    if out_dir != '' and (not os.path.exists(out_dir)):
        os.makedirs(out_dir)

    if use_plink:
        computeCovarianceMatrixPlink(
            plink_path, out_dir, bfile, cfile, sim_type=sim_type)
    else:
        computeCovarianceMatrixPython(out_dir, bfile, cfile, sim_type=sim_type)


def eighCovarianceMatrix(cfile):
    """
    compute similarity matrix using plink

    Input:
    cfile        :   the covariance matrix will be read from cfile.cov while the eigenvalues and the eigenverctors will
                        be written to cfile.cov.eval and cfile.cov.evec respectively
    """
    # precompute eigenvalue decomposition
    K = np.loadtxt(cfile + '.cov')
    K += 1e-4 * sp.eye(K.shape[0])
    S, U = la.eigh(K)
    S = S[::-1]
    U = U[:, ::-1]
    np.savetxt(cfile + '.cov.eval', S, fmt='%.6f')
    np.savetxt(cfile + '.cov.evec', U, fmt='%.6f')


def fit_null(Y, S_XX, U_XX, nfile, F):
    """
    fit null model

    Y       NxP phenotype matrix
    S_XX    eigenvalues of the relatedness matrix
    U_XX    eigen vectors of the relatedness matrix
    """
    mtSet = limix.MTSet(Y, S_R=S_XX, U_R=U_XX, F=F)

    RV = mtSet.fitNull(cache=False)
    params = np.array([RV['params0_g'], RV['params0_n']])
    np.savetxt(nfile + '.p0', params)
    np.savetxt(nfile + '.nll0', RV['NLL0'])
    np.savetxt(nfile + '.cg0', RV['Cg'])
    np.savetxt(nfile + '.cn0', RV['Cn'])
    # if F is not None: np.savetxt(nfile+'.f0',RV['params_mean'])


def preprocess(options):
    import matplotlib
    matplotlib.use('PDF')
    import pylab as pl

    assert options.bfile is not None, 'Please specify a bfile.'

    """ computing the covariance matrix """
    if options.compute_cov:
        assert options.bfile is not None, 'Please specify a bfile.'
        assert options.cfile is not None, 'Specify covariance matrix basename'
        print('Computing covariance matrix')
        t0 = time.time()
        computeCovarianceMatrix(
            options.plink_path,
            options.bfile,
            options.cfile,
            options.sim_type)
        t1 = time.time()
        print(('... finished in %s seconds' % (t1 - t0)))
        print('Computing eigenvalue decomposition')
        t0 = time.time()
        eighCovarianceMatrix(options.cfile)
        t1 = time.time()
        print(('... finished in %s seconds' % (t1 - t0)))

    """ computing principal components """
    if options.compute_PCs > 0:
        assert options.ffile is not None, 'Specify fix effects basename for saving PCs'
        t0 = time.time()
        computePCs(
            options.plink_path,
            options.compute_PCs,
            options.bfile,
            options.ffile)
        t1 = time.time()
        print(('... finished in %s seconds' % (t1 - t0)))

    """ fitting the null model """
    if options.fit_null:
        if options.nfile is None:
            options.nfile = os.path.split(options.bfile)[-1]
            warnings.warn('nfile not specifed, set to %s' % options.nfile)
        print('Fitting null model')
        assert options.pfile is not None, 'phenotype file needs to be specified'
        # read pheno
        Y = readPhenoFile(options.pfile, idx=options.trait_idx)
        # read covariance
        if options.cfile is None:
            cov = {'eval': None, 'evec': None}
            warnings.warn(
                'cfile not specifed, a one variance compoenent model will be considered')
        else:
            cov = readCovarianceMatrixFile(options.cfile, readCov=False)
            assert Y.shape[0] == cov['eval'].shape[0], 'dimension mismatch'
        # read covariates
        F = None
        if options.ffile is not None:
            F = readCovariatesFile(options.ffile)
            assert Y.shape[0] == F.shape[0], 'dimensions mismatch'
        t0 = time.time()
        fit_null(Y, cov['eval'], cov['evec'], options.nfile, F)
        t1 = time.time()
        print(('.. finished in %s seconds' % (t1 - t0)))



from numpy import asarray


def plot_kinship(K):
    r"""Plot Kinship matrix.

    Parameters
    ----------
    K : array_like
        Kinship matrix.

    Examples
    --------

    .. plot::

        from limix.io.npy import see_kinship
        from limix.io.examples import numpy_kinship_file_example

        see_kinship(numpy_kinship_file_example())
    """
    from matplotlib import pyplot as plt
    from matplotlib import cm as cm

    K = asarray(K, float)
    mi = K.min()
    ma = K.max()
    K = (K - mi) / (ma - mi)

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    cax = ax1.imshow(K, interpolation='nearest')
    fig.colorbar(cax)
    plt.show()



import numpy as np
import scipy.stats as st
from numpy import mean as _mean
from numpy import std as _std
from numpy import arange


def draw_normal(axis, mean, scale, nstd, color):
    max_pdf = st.norm.pdf(mean, mean, scale)

    axis.plot([mean, mean], [0, max_pdf], color=color, linestyle="--")

    axis.annotate(
        '$\mu$',
        xy=(mean + 0.6 * scale, max_pdf),
        horizontalalignment='center',
        verticalalignment='bottom',
        fontsize=15,
        color=color)

    top = st.norm.pdf(mean + nstd * scale, mean, scale)
    left = mean - nstd * scale
    right = mean + nstd * scale

    axis.plot([right, right], [0, top], color=color, linestyle="--")

    axis.plot([left, left], [0, top], color=color, linestyle="--")

    if int(nstd) == nstd:
        mu_sigma = '$\mu+%d\sigma$' % nstd
    else:
        mu_sigma = '$\mu+%.1f\sigma$' % nstd

    axis.annotate(
        mu_sigma,
        xy=(mean + (1.2 + nstd) * scale, top),
        horizontalalignment='center',
        verticalalignment='bottom',
        fontsize=15,
        color=color)


def plot_normal(x, bins=20, nstd=2, figure=None):
    r"""Plot a fit of a normal distribution to the data in x.

    Args:
        x (array_like): values to be fitted.
        bins (int): number of histogram bins.
        nstd (float): standard deviation multiplier.
        figure (:class:`matplotlib.figure.Figure`): matplotlib figure or
                                                    ``None``.

    Returns:
        :class:`matplotlib.figure.Figure`: figure to be shown.

    Examples
    --------

        .. plot::

            from numpy.random import RandomState
            from limix.plot import plot_normal
            from matplotlib import pyplot as plt
            random = RandomState(10)
            x = random.randn(100)
            f = plot_normal(x, nstd=2)
            plt.show(f)
    """
    import matplotlib.pyplot as plt

    if figure is None:
        figure = plt.figure()

    ax = figure.gca()

    mean_x = _mean(x)
    std_x = _std(x)

    xvals = arange(mean_x - 5 * std_x, mean_x + 5 * std_x, .001)
    yvals = st.norm.pdf(xvals, mean_x, std_x)

    ax.hist(x, bins, normed=True)

    ax.plot(xvals, yvals, color='red')

    draw_normal(ax, mean_x, std_x, nstd, 'red')

    return figure



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import numpy as np
import scipy as sp
import scipy.stats as st


class LMM:
    r"""Basic class for univariate single-variant association testing with LMMs.

    Args:
        snps (ndarray):
            (`N`, `S`) ndarray of `S` SNPs for `N` individuals.
        pheno (ndarray):
            (`N`, `P`) ndarray of `P` phenotype sfor `N` individuals.
            If phenotypes have missing values, then the subset of
            individuals used for each phenotype column will be subsetted.
        K (ndarray, optional):
            (`N`, `N`) ndarray of LMM-covariance/kinship coefficients (optional)
            If not provided, then standard linear regression is considered.
        covs (ndarray, optional):
            (`N`, `D`) ndarray of `D` covariates for `N` individuals.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        test ({'lrt', 'f'}, optional):
            test statistic.
            'lrt' for likelihood ratio test (default) or 'f' for F-test.
        NumIntervalsDelta0 (int, optional):
            number of steps for delta optimization on the null model.
            By default ``NumIntervalsDelta0`` is 100.
        NumIntervalsDeltaAlt (int, optional):
            number of steps for delta optimization on the alternative model.
            Requires ``searchDelta=True`` to have an effect.
        searchDelta (bool, optional):
            if True, delta optimization on the alternative model is carried out.
            By default ``searchDelta`` is False.
        verbose (bool, optional):
            if True, details such as runtime as displayed.

    Examples
    --------

        Example of single-variant association testing using a linear mixed model.

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.qtl import LMM
            >>> from numpy import dot
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>> random = RandomState(1)
            >>>
            >>> N = 100
            >>> S = 1000
            >>>
            >>> # generate data
            >>> snps = (random.rand(N, S) < 0.2).astype(float)
            >>> pheno = random.randn(N, 1)
            >>> W = random.randn(N, 10)
            >>> kinship = dot(W, W.T) / float(10)
            >>>
            >>> # run single-variant associaiton testing with LMM
            >>> lmm = LMM(snps, pheno, kinship)
            >>> pv = lmm.getPv()
            >>> beta = lmm.getBetaSNP()
            >>> beta_ste = lmm.getBetaSNPste()
            >>>
            >>> print(pv.shape)
            (1, 1000)
            >>> print(pv[:,:4]) # doctest: +SKIP
            [[0.8571 0.4668 0.5872 0.5589]]
            >>> print(beta[:,:4]) # doctest: +SKIP
            [[ 0.0436 -0.1695 -0.12    0.1388]]
            >>> print(beta_ste[:,:4]) # doctest: +SKIP
            [[0.2422 0.2329 0.221  0.2375]]

        As shown in the exmaple below,
        the method can be applied directly on multiple phenotypes.
        Each phenotype is indipendently tested against all variants one-by-one.

        .. doctest::

            >>> random = RandomState(1)
            >>>
            >>> P = 3
            >>> phenos = random.randn(N, P)
            >>>
            >>> lmm = LMM(snps, phenos, kinship)
            >>> pv = lmm.getPv()
            >>>
            >>> print(pv.shape)
            (3, 1000)
            >>> print(pv[:,:4]) # doctest: +SKIP
            [[0.4712 0.694  0.3369 0.5221]
             [0.7336 0.8799 0.8109 0.1071]
             [0.0662 0.9203 0.2873 0.8268]]
    """

    def __init__(self,
                 snps,
                 pheno,
                 K=None,
                 covs=None,
                 test='lrt',
                 NumIntervalsDelta0=100,
                 NumIntervalsDeltaAlt=100,
                 searchDelta=False,
                 verbose=None):
        # create column of 1 for fixed if nothing provide
        import limix_legacy.deprecated
        import limix_legacy.deprecated as dlimix_legacy

        if len(pheno.shape) == 1:
            pheno = pheno[:, sp.newaxis]

        self.verbose = dlimix_legacy.getVerbose(verbose)
        self.snps = snps
        self.pheno = pheno
        self.K = K
        self.covs = covs
        self.test = test
        self.NumIntervalsDelta0 = NumIntervalsDelta0
        self.NumIntervalsDeltaAlt = NumIntervalsDeltaAlt
        self.searchDelta = searchDelta
        self.verbose = verbose
        self.N = self.pheno.shape[0]
        self.P = self.pheno.shape[1]
        self.Iok = ~(np.isnan(self.pheno).any(axis=1))
        if self.K is None:
            self.searchDelta = False
            self.K = np.eye(self.snps.shape[0])
        if self.covs is None:
            self.covs = np.ones((self.snps.shape[0], 1))

        self._lmm = None
        # run
        self.verbose = verbose
        self.process()

    def process(self):

        import limix_legacy.deprecated
        import limix_legacy.deprecated as dlimix_legacy

        t0 = time.time()
        if self._lmm is None:
            self._lmm = limix_legacy.deprecated.CLMM()
            self._lmm.setK(self.K)
            self._lmm.setSNPs(self.snps)
            self._lmm.setPheno(self.pheno)
            self._lmm.setCovs(self.covs)
            if self.test == 'lrt':
                self._lmm.setTestStatistics(self._lmm.TEST_LRT)
            elif self.test == 'f':
                self._lmm.setTestStatistics(self._lmm.TEST_F)
            else:
                print((self.test))
                raise NotImplementedError("only f and lrt are implemented")
            # set number of delta grid optimizations?
            self._lmm.setNumIntervals0(self.NumIntervalsDelta0)
            if self.searchDelta:
                self._lmm.setNumIntervalsAlt(self.NumIntervalsDeltaAlt)
            else:
                self._lmm.setNumIntervalsAlt(0)

        if not np.isnan(self.pheno).any():
            # process
            self._lmm.process()
            self.pvalues = self._lmm.getPv()
            self.beta_snp = self._lmm.getBetaSNP()
            self.beta_ste = self._lmm.getBetaSNPste()
            self.ldelta_0 = self._lmm.getLdelta0()
            self.ldelta_alt = self._lmm.getLdeltaAlt()
            self.NLL_0 = self._lmm.getNLL0()
            self.NLL_alt = self._lmm.getNLLAlt()
        else:
            if self._lmm is not None:
                raise Exception(
                    'cannot reuse a CLMM object if missing variables are present'
                )
            else:
                self._lmm = limix_legacy.deprecated.CLMM()
            # test all phenotypes separately
            self.pvalues = np.zeros((self.phenos.shape[1], self.snps.shape[1]))
            self.beta_snp = np.zeros((self.phenos.shape[1],
                                      self.snps.shape[1]))
            self.beta_ste = np.zeros((self.phenos.shape[1],
                                      self.snps.shape[1]))
            self.ldelta_0 = np.zeros((self.phenos.shape[1],
                                      self.snps.shape[1]))
            self.ldelta_alt = np.zeros((self.phenos.shape[1],
                                        self.snps.shape[1]))
            self.NLL_0 = np.zeros((self.phenos.shape[1], self.snps.shape[1]))
            self.NLL_alt = np.zeros((self.phenos.shape[1], self.snps.shape[1]))
            self.test_statistics = np.zeros((self.phenos.shape[1],
                                             self.snps.shape[1]))
            for ip in np.arange(self.phenos.shape[1]):
                pheno_ = self.phenos[:, ip]
                i_nonz = ~(pheno_.isnan())

                self._lmm.setK(self.K[i_nonz, i_nonz])
                self._lmm.setSNPs(self.snps[i_nonz])
                self._lmm.setPheno(pheno_[i_nonz, np.newaxis])
                self._lmm.setCovs(self.covs[i_nonz])
                self._lmm.process()
                self.pvalues[ip:ip + 1] = self._lmm.getPv()
                self.beta_snp[ip:ip + 1] = self._lmm.getBetaSNP()
                self.beta_ste[ip:ip + 1] = self._lmm.getBetaSNPste()
                self.ldelta_0[ip:ip + 1] = self._lmm.getLdelta0()
                self.ldelta_alt[ip:ip + 1] = self._lmm.getLdeltaAlt()
                self.NLL_0[ip:ip + 1] = self._lmm.getNLL0()
                self.NLL_alt[ip:ip + 1] = self._lmm.getNLLAlt()
                self.test_statistics[ip:ip + 1] = self._lmm.getTestStatistics()
                pass
        if self._lmm.getTestStatistics(
        ) == self._lmm.TEST_LRT and self.test != "lrt":
            raise NotImplementedError("only f and lrt are implemented")
        elif self._lmm.getTestStatistics(
        ) == self._lmm.TEST_F and self.test != "f":
            raise NotImplementedError("only f and lrt are implemented")

        if self._lmm.getTestStatistics() == self._lmm.TEST_F:
            self.test_statistics = (self.beta_snp * self.beta_snp) / (
                self.beta_ste * self.beta_ste)
        if self._lmm.getTestStatistics() == self._lmm.TEST_LRT:
            self.test_statistics = 2.0 * (self.NLL_0 - self.NLL_alt)
        t1 = time.time()

        if self.verbose:
            print(("finished GWAS testing in %.2f seconds" % (t1 - t0)))

    def setCovs(self, covs):
        self._lmm.setCovs(covs)

    def getBetaSNP(self):
        """
        Returns:
            ndarray: (`P`, `S`) ndarray of SNP effect sizes.
        """
        return self.beta_snp

    def getPv(self):
        """
        Returns:
            ndarray: (`P`, `S`) ndarray of P-values.
        """
        return self.pvalues

    def getBetaSNPste(self):
        """
        Returns:
            ndarray: (`P`, `S`) ndarray of standard errors over SNP effects.
        """
        beta = self.getBetaSNP()
        pv = self.getPv()
        z = sp.sign(beta) * sp.sqrt(st.chi2(1).isf(pv))
        ste = beta / z
        return ste



from optparse import OptionParser

from ..mtset.core.definesetsCore import definesets


def entry_point():
    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    parser.add_option("--wfile", dest='wfile', type=str, default=None)

    parser.add_option(
        "--sliding_window",
        action="store_true",
        dest="sliding_wind",
        default=False)

    parser.add_option(
        "--filter_sets",
        action="store_true",
        dest='filter_sets',
        default=False)

    parser.add_option(
        "--window_size", dest='window_size', type=int, default=3e4)

    parser.add_option("--step", dest='step', type=int, default=15e3)

    parser.add_option("--minSnps", dest='minSnps', type=int, default=1)
    parser.add_option("--maxSnps", dest='maxSnps', type=int, default=None)
    parser.add_option("--chrom", dest='chrom', type=int, default=None)

    parser.add_option("--iwfile", dest='iwfile', type=str, default=None)

    (options, args) = parser.parse_args()

    definesets(options)



from argparse import ArgumentParser


def do_see(args):
    import limix

    if args.type is None:
        ft = limix.io.file_type(args.file)
    else:
        ft = args.type

    if ft == 'hdf5':
        limix.io.hdf5.see(args.file, show_chunks=args.show_chunks)
    elif ft == 'csv':
        limix.io.csv.see(args.file)
    elif ft == 'grm.raw':
        limix.io.plink.see_kinship(args.file)
    elif ft == 'npy':
        limix.io.npy.see_kinship(args.file)
    elif ft == 'bed':
        limix.io.plink.see_bed(args.file)
    elif ft == 'image':
        limix.plot.see_image(args.file)
    else:
        print("Unknown file type: %s" % args.file)


def see_parser(parser):
    parser.add_argument('file', help='file path')
    parser.add_argument(
        '--show-chunks',
        dest='show_chunks',
        action='store_true',
        help='show chunk information for hdf5 files')
    parser.add_argument('--type', dest='type', help='specify file type')
    parser.set_defaults(show_chunks=False)
    parser.set_defaults(func=do_see)
    return parser


def entry_point():
    p = ArgumentParser()

    subparsers = p.add_subparsers(title='subcommands')
    see_parser(subparsers.add_parser('see'))

    args = p.parse_args()
    if hasattr(args, 'func'):
        func = args.func
        del args.func
        func(args)
    else:
        p.print_help()



from optparse import OptionParser


def entry_point():
    from ..mtset.core.preprocessCore import preprocess

    parser = OptionParser()
    parser.add_option("--bfile", dest='bfile', type=str, default=None)
    parser.add_option("--cfile", dest='cfile', type=str, default=None)
    parser.add_option("--pfile", dest='pfile', type=str, default=None)
    parser.add_option("--nfile", dest='nfile', type=str, default=None)
    parser.add_option("--ffile", dest='ffile', type=str, default=None)
    parser.add_option("--trait_idx", dest='trait_idx', type=str, default=None)

    parser.add_option(
        "--compute_covariance",
        action="store_true",
        dest="compute_cov",
        default=False)
    parser.add_option("--compute_PCs", dest="compute_PCs", default=0, type=int)

    parser.add_option(
        "--plink_path", dest='plink_path', type=str, default='plink')
    parser.add_option("--sim_type", dest='sim_type', type=str, default='RRM')

    parser.add_option(
        "--fit_null", action="store_true", dest="fit_null", default=False)

    (options, args) = parser.parse_args()

    preprocess(options)



"""
Created on Sep 24, 2013
@author: casale
"""
import scipy as SP
import scipy.linalg as LA
import pdb

import sys
from . import plink_reader


def genBinormal(dim1, dim2, percSign=0.5, std=1e-1):
    rv = (2 * (SP.rand(dim1, dim2) > percSign) - 1) + \
        std * SP.randn(dim1, dim2)
    return rv


def selectRnd(n_sel, n_all):
    rv = SP.zeros(n_all)
    rv[:n_sel] = 1
    SP.random.shuffle(rv)
    rv = rv == 1
    return rv


class CSimulator:
    """
    this class takes care of phenotype generation in a flexible way
    """

    def __init__(
            self,
            bfile=None,
            XX=None,
            X=None,
            chrom=None,
            pos=None,
            P=1,
            trait_effect='shared'):
        """
        X: genotype matrix
        traitNum: number of traits to be considered
        """
        self.bfile = bfile
        self.N = XX.shape[0]
        self.P = P
        self.XX = XX

        self.X = X
        self.chrom = chrom
        self.pos = pos

        self.trait_effect = trait_effect

    def setTraitEffect(self, trait_effect):
        self.trait_effect = trait_effect

    def getRegion(
            self,
            size=3e4,
            min_nSNPs=1,
            chrom_i=None,
            pos_min=None,
            pos_max=None):
        """
        Sample a region from the piece of genotype X, chrom, pos
        minSNPnum:  minimum number of SNPs contained in the region
        Ichrom:  restrict X to chromosome Ichrom before taking the region
        cis:        bool vector that marks the sorted region
        region:  vector that contains chrom and init and final position of the region
        """
        if (self.chrom is None) or (self.pos is None):
            bim = plink_reader.readBIM(self.bfile, usecols=(0, 1, 2, 3))
            chrom = SP.array(bim[:, 0], dtype=int)
            pos = SP.array(bim[:, 3], dtype=int)
        else:
            chrom = self.chrom
            pos = self.pos

        if chrom_i is None:
            n_chroms = chrom.max()
            chrom_i = int(SP.ceil(SP.rand() * n_chroms))

        pos = pos[chrom == chrom_i]
        chrom = chrom[chrom == chrom_i]

        ipos = SP.ones(len(pos), dtype=bool)
        if pos_min is not None:
            ipos = SP.logical_and(ipos, pos_min < pos)

        if pos_max is not None:
            ipos = SP.logical_and(ipos, pos < pos_max)

        pos = pos[ipos]
        chrom = chrom[ipos]

        if size == 1:
            # select single SNP
            idx = int(SP.ceil(pos.shape[0] * SP.rand()))
            cis = SP.arange(pos.shape[0]) == idx
            region = SP.array([chrom_i, pos[idx], pos[idx]])
        else:
            while True:
                idx = int(SP.floor(pos.shape[0] * SP.rand()))
                posT1 = pos[idx]
                posT2 = pos[idx] + size
                if posT2 <= pos.max():
                    cis = chrom == chrom_i
                    cis *= (pos > posT1) * (pos < posT2)
                    if cis.sum() > min_nSNPs:
                        break
            region = SP.array([chrom_i, posT1, posT2])

        start = SP.nonzero(cis)[0].min()
        nSNPs = cis.sum()

        if self.X is None:
            rv = plink_reader.readBED(
                self.bfile,
                useMAFencoding=True,
                start=start,
                nSNPs=nSNPs,
                bim=bim)
            Xr = rv['snps']
        else:
            Xr = self.X[:, start:start + nSnps]

        return Xr, region

    def genRegionTerm(
            self,
            X,
            vTot=0.1,
            pCausal=0.10,
            nCausal=None,
            pCommon=1.,
            nCommon=None,
            plot=False,
            distribution='biNormal'):
        """
        Generate population structure term
        Population structure is simulated by background SNPs

        beta_pdf:        pdf used to generate the regression weights
                          for now either Normal or fixed
        variance:        variance of the term
        percCausal:    percentage of causal SNPs
        Xcausal:          set of SNPs being causal
        """
        S = X.shape[1]

        # number of causal, common, specific
        if nCausal is None:
            nCausal = int(SP.floor(pCausal * S))
        if nCommon is None:
            nCommon = round(pCommon * nCausal)
        nSpecific = self.P * (nCausal - nCommon)

        # common SNPs
        if nCommon > 0:
            if distribution == 'biNormal':
                Bc = SP.dot(
                    genBinormal(
                        nCommon, self.P), self.genTraitEffect(
                        distribution=distribution))
            elif distribution == 'normal':
                Bc = SP.dot(
                    self.genWeights(
                        nCommon, self.P), self.genTraitEffect(
                        distribution=distribution))
            Ic = selectRnd(nCommon, S)
            Yc = SP.dot(X[:, Ic], Bc)
            Yc *= SP.sqrt(nCommon / Yc.var(0).mean())
        else:

            Yc = SP.zeros((self.N, self.P))

        # indipendent signal
        if nSpecific > 0:
            Is = selectRnd(nSpecific, S * self.P).reshape(S, self.P)
            if distribution == 'biNormal':
                Bi = Is * genBinormal(S, self.P)
            elif distribution == 'normal':
                Bi = Is * SP.randn(S, self.P)
            Yi = SP.dot(X, Bi)
            Yi *= SP.sqrt(nSpecific / (Yi.var(0).mean() * self.P))
        else:
            Yi = SP.zeros((self.N, self.P))

        Y = Yc + Yi
        Yc *= SP.sqrt(vTot / Y.var(0).mean())
        Yi *= SP.sqrt(vTot / Y.var(0).mean())

        if plot:
            import pylab as PL
            PL.ion()
            for p in range(self.P):
                PL.subplot(self.P, 1, p + 1)
                PL.plot(SP.arange(S)[Ic], Bc[:, p], 'o', color='y')
                _Is = Is[:, p]
                if _Is.sum() > 0:
                    PL.plot(SP.arange(S)[_Is], Bi[_Is, p], 'o', color='r')
                # PL.ylim(-2,2)
                PL.plot([0, S], [0, 0], 'k')

        return Yc, Yi

    def _genBgTerm_fromSNPs(
            self,
            vTot=0.5,
            vCommon=0.1,
            pCausal=0.5,
            plot=False):
        """ generate  """

        if self.X is None:
            print('Reading in all SNPs. This is slow.')
            rv = plink_reader.readBED(self.bfile, useMAFencoding=True)
            X = rv['snps']
        else:
            X = self.X

        S = X.shape[1]
        vSpecific = vTot - vCommon

        # select causal SNPs
        nCausal = int(SP.floor(pCausal * S))
        Ic = selectRnd(nCausal, S)
        X = X[:, Ic]

        # common effect
        Bc = SP.dot(self.genWeights(nCausal, self.P), self.genTraitEffect())
        Yc = SP.dot(X, Bc)
        Yc *= SP.sqrt(vCommon / Yc.var(0).mean())

        # indipendent effect
        Bi = SP.randn(nCausal, self.P)
        Yi = SP.dot(X, Bi)
        Yi *= SP.sqrt(vSpecific / Yi.var(0).mean())

        if plot:
            import pylab as PL
            PL.ion()
            for p in range(self.P):
                PL.subplot(self.P, 1, p + 1)
                PL.plot(SP.arange(self.X.shape[1])[
                        Ic], Bc[:, p], 'o', color='y', alpha=0.05)
                PL.plot(SP.arange(self.X.shape[1])[
                        Ic], Bi[:, p], 'o', color='r', alpha=0.05)
                # PL.ylim(-2,2)
                PL.plot([0, Ic.shape[0]], [0, 0], 'k')

        return Yc, Yi

    def _genBgTerm_fromXX(self, vTot, vCommon, XX, a=None, c=None):
        """
        generate background term from SNPs

        Args:
            vTot: variance of Yc+Yi
            vCommon: variance of Yc
            XX: kinship matrix
            a: common scales, it can be set for debugging purposes
            c: indipendent scales, it can be set for debugging purposes
        """
        vSpecific = vTot - vCommon

        SP.random.seed(0)
        if c is None:
            c = SP.randn(self.P)
        XX += 1e-3 * SP.eye(XX.shape[0])
        L = LA.cholesky(XX, lower=True)

        # common effect
        R = self.genWeights(self.N, self.P)
        A = self.genTraitEffect()
        if a is not None:
            A[0, :] = a
        Yc = SP.dot(L, SP.dot(R, A))
        Yc *= SP.sqrt(vCommon) / SP.sqrt(Yc.var(0).mean())

        # specific effect
        R = SP.randn(self.N, self.P)
        Yi = SP.dot(L, SP.dot(R, SP.diag(c)))
        Yi *= SP.sqrt(vSpecific) / SP.sqrt(Yi.var(0).mean())

        return Yc, Yi

    def genBgTerm(
            self,
            vTot=0.5,
            vCommon=0.1,
            pCausal=0.5,
            XX=None,
            use_XX=False,
            a=None,
            c=None,
            plot=False):
        """ generate  """
        if use_XX:
            if XX is None:
                XX = self.XX
            assert XX is not None, 'Simulator: set XX!'
            Yc, Yi = self._genBgTerm_fromXX(vTot, vCommon, XX, a=a, c=c)
        else:
            Yc, Yi = self._genBgTerm_fromSNPs(
                vTot=vTot, vCommon=vCommon, pCausal=pCausal, plot=plot)
        return Yc, Yi

    def genHidden(self, nHidden=10, vTot=0.5, vCommon=0.1):
        """ generate  """

        vSpecific = vTot - vCommon

        # generate hidden
        X = self.genWeights(self.N, nHidden)

        # common effect
        H = self.genWeights(nHidden, self.P)
        Bc = SP.dot(H, self.genTraitEffect())
        Yc = SP.dot(X, Bc)
        Yc *= SP.sqrt(vCommon / Yc.var(0).mean())

        # indipendent effect
        Bi = SP.randn(nHidden, self.P)
        Yi = SP.dot(X, Bi)
        Yi *= SP.sqrt(vSpecific / Yi.var(0).mean())

        return Yc, Yi

    def genTraitEffect(self, distribution='normal'):

        W = SP.zeros((self.P, self.P))
        if self.trait_effect == 'shared':

            if distribution == 'normal':
                W[0, :] = SP.random.randn(1, self.P)
            else:
                W[0, :] = genBinormal(1, self.P)

        elif self.trait_effect == 'tanh':
            assert distribution == 'normal', 'tanh trait effect is only implemented for normal distributed effects'
            X = 10 * SP.linspace(0, 1, self.P) - 5
            a = - SP.absolute(SP.random.randn())
            c = - SP.absolute(SP.random.randn())
            Y = SP.tanh(a * X + c)
            W[0, :] = Y

        return W

    def genWeights(self, N, P):
        if self.trait_effect == 'shared':
            return SP.random.randn(N, P)

        elif self.trait_effect == 'tanh':
            # ensure that weights are positive...
            return SP.absolute(SP.random.randn(N, P))

    def genNoise(self, vTot=0.4, vCommon=0.2):

        vSpecifc = vTot - vCommon

        # common
        Yc = SP.dot(self.genWeights(self.N, self.P), self.genTraitEffect())
        Yc *= SP.sqrt(vCommon / Yc.var(0).mean())

        # independent
        Yi = SP.randn(self.N, self.P)
        Yi *= SP.sqrt(vSpecifc / Yi.var(0).mean())

        return Yc, Yi

    def genPheno(
            self,
            Xr,
            vTotR=0.1,
            nCommonR=5,
            nCausalR=10,
            distribution='biNormal',
            vCommonBg=0.1,
            vTotBg=0.4,
            pCausalBg=0.5,
            XX=None,
            use_XX=False,
            vCommonH=0.1,
            vTotH=0.2,
            nHidden=10,
            vCommonN=0.,
            vTotN=0.3,
            standardize=True):

        YRc, YRi = self.genRegionTerm(
            Xr, vTot=vTotR, nCommon=nCommonR, nCausal=nCausalR, distribution=distribution)
        YGc, YGi = self.genBgTerm(
            vCommon=vCommonBg, vTot=vTotBg, pCausal=pCausalBg, XX=XX, use_XX=use_XX)
        YHc, YHi = self.genHidden(
            vCommon=vCommonH, vTot=vTotH, nHidden=nHidden)
        YNc, YNi = self.genNoise(vCommon=vCommonN, vTot=vTotN)
        Y = YRc + YRi + YGc + YGi + YHc + YHi + YNc + YNi

        if standardize:
            Y -= Y.mean(0)
            Y /= Y.std(0)

        info = {'YRc': YRc, 'YRi': YRi,
                'YGc': YGc, 'YGi': YGi,
                'YHc': YHc, 'YHi': YHi,
                'YNc': YNc, 'YNi': YNi}

        return Y, info



r"""
***************************
Interaction Set Test (iSet)
***************************

Joint genetic models for multiple traits have helped to enhance association
analyses.
Most existing multi-trait models have been designed to increase power for
detecting associations, whereas the analysis of interactions has received
considerably less attention.
Here, we implement :func:`.fit_iSet` [CHRS17]_, a method based on linear
mixed models to test for interactions between sets of variants and
environmental states or other contexts.
Our model generalizes previous
interaction tests and in particular provides a test for local differences in
the genetic architecture between contexts.
We first use simulations to validate
iSet before applying the model to the analysis of genotype-environment
interactions in an eQTL study.
Our model retrieves a larger number of
interactions than alternative methods and reveals that up to 20%% of cases show
context-specific configurations of causal variants.
Finally, we apply iSet to
test for sub-group specific genetic effects in human lipid levels in a large
human cohort, where we identify a gene-sex interaction for C-reactive protein
that is missed by alternative methods.

.. rubric:: References

.. [CHRS17] Casale FP, Horta D, Rakitsch B, Stegle O (2017) Joint genetic analysis using variant sets reveals polygenic gene-context interactions. PLOS Genetics 13(4): e1006693.

Public interface
^^^^^^^^^^^^^^^^
"""

from .iset import fit_iSet

__all__ = ['fit_iSet']



from optparse import OptionParser
import scipy as SP

import os
from . import simulator as sim
from .read_utils import readCovarianceMatrixFile
from .read_utils import readBimFile


def genPhenoCube(
        sim,
        Xr,
        vTotR=4e-3,
        nCausalR=10,
        pCommonR=0.8,
        vTotBg=0.4,
        pHidd=0.6,
        pCommon=0.8):
    # region
    nCommonR = int(SP.around(nCausalR * pCommonR))
    # background
    vCommonBg = pCommon * vTotBg
    # noise
    vTotH = pHidd * (1 - vTotR - vTotBg)
    vTotN = (1 - pHidd) * (1 - vTotR - vTotBg)
    vCommonH = pCommon * vTotH
    all_settings = {
        'vTotR': vTotR,
        'nCommonR': nCommonR,
        'nCausalR': nCausalR,
        'vTotBg': vTotBg,
        'vCommonBg': vCommonBg,
        'pCausalBg': 1.,
        'use_XX': True,
        'vTotH': vTotH,
        'vCommonH': vCommonH,
        'nHidden': 10,
        'vTotN': vTotN,
        'vCommonN': 0.}

    Y, info = sim.genPheno(Xr, **all_settings)
    return Y, info


def simPheno(options):

    print('importing covariance matrix')
    if options.cfile is None:
        options.cfile = options.bfile
    XX = readCovarianceMatrixFile(options.cfile, readEig=False)['K']

    print('simulating phenotypes')
    SP.random.seed(options.seed)
    simulator = sim.CSimulator(bfile=options.bfile, XX=XX, P=options.nTraits)
    Xr, region = simulator.getRegion(chrom_i=options.chrom, size=options.windowSize,
                                     min_nSNPs=options.nCausalR, pos_min=options.pos_min, pos_max=options.pos_max)

    Y, info = genPhenoCube(simulator, Xr, vTotR=options.vTotR, nCausalR=options.nCausalR,
                           pCommonR=options.pCommonR, vTotBg=options.vTotBg, pHidd=options.pHidden, pCommon=options.pCommon)

    print('exporting pheno file')
    if options.pfile is not None:
        outdir = os.path.split(options.pfile)[0]
        if not os.path.exists(outdir):
            os.makedirs(outdir)
    else:
        identifier = '_seed%d_nTraits%d_wndSize%d_vTotR%.2f_nCausalR%d_pCommonR%.2f_vTotBg%.2f_pHidden%.2f_pCommon%.2f' % (
            options.seed, options.nTraits, options.windowSize, options.vTotR, options.nCausalR, options.pCommonR, options.vTotBg, options.pHidden, options.pCommon)
        options.pfile = os.path.split(options.bfile)[-1] + '%s' % identifier

    pfile = options.pfile + '.phe'
    rfile = options.pfile + '.phe.region'

    SP.savetxt(pfile, Y)
    SP.savetxt(rfile, region)



from __future__ import division

from time import time

from numpy import asarray, diag, ones

from glimix_core.glmm import GLMM
from limix.qtl.lmm import LMM
from numpy_sugar.linalg import economic_qs


def qtl_test_glmm(snps,
                  pheno,
                  lik,
                  K,
                  covs=None,
                  test='lrt',
                  NumIntervalsDeltaAlt=100,
                  searchDelta=False,
                  verbose=True):
    """
    Wrapper function for univariate single-variant association testing
    using a generalised linear mixed model.

    Args:
        snps (array_like):
            `N` individuals by `S` SNPs.
        pheno (tuple, array_like):
            Either a tuple of two arrays of `N` individuals each (Binomial
            phenotypes) or an array of `N` individuals (Poisson or Bernoulli
            phenotypes). It does not support missing values yet.
        lik ({'bernoulli', 'binomial', 'poisson'}):
            Sample likelihood describing the residual distribution.
        K (array_like):
            `N` by `N` covariance matrix (e.g., kinship coefficients).
        covs (array_like, optional):
            `N` individuals by `D` covariates.
            By default, ``covs`` is a (`N`, `1`) array of ones.
        test ({'lrt'}, optional):
            Likelihood ratio test (default).
        NumIntervalsDeltaAlt (int, optional):
            number of steps for delta optimization on the alternative model.
            Requires ``searchDelta=True`` to have an effect.
        searchDelta (bool, optional):
            if ``True``, delta optimization on the alternative model is
            carried out. By default ``searchDelta`` is ``False``.
        verbose (bool, optional):
            if ``True``, details such as runtime are displayed.

    Returns:
        :class:`limix.qtl.LMM`: LIMIX LMM object

    Examples
    --------
    .. doctest::

        >>> from numpy import dot, exp, sqrt
        >>> from numpy.random import RandomState
        >>> from limix.qtl import qtl_test_glmm
        >>>
        >>> random = RandomState(0)
        >>>
        >>> G = random.randn(250, 500) / sqrt(500)
        >>> beta = 0.01 * random.randn(500)
        >>>
        >>> z = dot(G, beta) + 0.1 * random.randn(250)
        >>> z += dot(G[:, 0], 1) # causal SNP
        >>>
        >>> y = random.poisson(exp(z))
        >>>
        >>> candidates = G[:, :5]
        >>> K = dot(G[:, 5:], G[:, 5:].T)
        >>> lm = qtl_test_glmm(candidates, y, 'poisson', K, verbose=False)
        >>>
        >>> print(lm.getPv()) # doctest: +SKIP
        [[0.0694 0.3336 0.5899 0.7388 0.7796]]
    """

    snps = _asarray(snps)

    if covs is None:
        covs = ones((snps.shape[0], 1))
    else:
        covs = _asarray(covs)

    K = _asarray(K)

    if isinstance(pheno, (tuple, list)):
        y = tuple([asarray(p, float) for p in pheno])
    else:
        y = asarray(pheno, float)

    start = time()
    QS = economic_qs(K)
    glmm = GLMM(y, lik, covs, QS)
    glmm.feed().maximize(verbose=verbose)
    # if verbose:
    #     print("Elapsed time for GLMM part: %.3f" % (time() - start))

    # extract stuff from glmm
    eta = glmm._site.eta
    tau = glmm._site.tau
    scale = float(glmm.scale)
    delta = float(glmm.delta)

    # define useful quantities
    mu = eta / tau
    var = 1. / tau
    s2_g = scale * (1 - delta)
    tR = s2_g * K + diag(var - var.min() + 1e-4)

    start = time()
    lmm = LMM(snps=snps, pheno=mu, K=tR, covs=covs, verbose=verbose)
    # if verbose:
    #     print("Elapsed time for LMM part: %.3f" % (time() - start))

    return lmm


def _asarray(X):
    import dask.array as da

    if not isinstance(X, da.Array):
        X = asarray(X, float)
    return X



import sys
import os
import subprocess
import pdb
import sys
import csv
import glob
import numpy as NP
from optparse import OptionParser
import time
import scipy as SP
import pandas as pd


def postprocess(options):
    r"""
    perform parametric fit of the test statistics and provide
    permutation and test pvalues
    """

    from limix.stats import Chi2mixture

    resdir = options.resdir
    out_file = options.outfile
    tol = options.tol

    print('.. load permutation results')
    file_name = os.path.join(resdir, 'perm*', '*.res')
    files = glob.glob(file_name)
    df0 = []
    for _file in files:
        print(_file)
        _df = pd.DataFrame.from_csv(_file, sep='\t', index_col=None)
        df0.append(_df)
    df0 = pd.concat(df0)

    print('.. fit test statistics')
    t0 = time.time()
    c2m = Chi2mixture(tol=4e-3)
    c2m.estimate_chi2mixture(df0['LLR'].values)
    t1 = time.time()
    print(('finished in %s seconds' % (t1 - t0)))

    print('.. load test results')
    file_name = os.path.join(resdir, 'test', '*.res')
    files = glob.glob(file_name)

    df = []
    for _file in files:
        print(_file)
        _df = pd.DataFrame.from_csv(_file, sep='\t', index_col=None)
        df.append(_df)
    df = pd.concat(df)

    print('.. calc pvalues')
    df0['pv'] = c2m.sf(df0['LLR'].values)
    df['pv'] = c2m.sf(df['LLR'].values)[:, NP.newaxis]

    print('.. export test results')
    perm_file = out_file + '.test'
    df.to_csv(perm_file, sep='\t', index=False)

    print('.. export permutation results')
    perm_file = out_file + '.perm'
    df0.to_csv(perm_file, sep='\t', index=False)



# def test_mtSet_simPheno_simulation(script_runner):
#     pass
#     # ret = script_runner.run(bfile='path_to_bfile', cfile='path_to_cfile')
#     # It is not returning anything as of now, but maybe we could return
#     # if it failed or not.
#
#     # Example:
#     # with open('output.txt', 'r') as f:
#     #     r = f.read()
#     #     assert r == 'expected output'



"""@package fdr
FDR estimation using Benjamini Hochberg and Stories method
"""

import logging as LG
import pdb
import pickle
import sys

import numpy as np
import scipy as sp
import scipy.interpolate
import scipy.stats as st


def qvalues1(PV, m=None, pi=1.0):
    """estimate q vlaues from a list of Pvalues
    this algorihm is taken from Storey, significance testing for genomic ...
    m: number of tests, (if not len(PV)), pi: fraction of expected true null (1.0 is a conservative estimate)
    @param PV: pvalues
    @param m:  total number of tests if PV is not the entire array.
    @param pi: fraction of true null
    """

    S = PV.shape
    PV = PV.flatten()
    if m is None:
        m = len(PV) * 1.0
    else:
        m *= 1.0
    lPV = len(PV)

    # 1. sort pvalues
    PV = PV.squeeze()
    IPV = PV.argsort()
    PV = PV[IPV]

    # 2. estimate lambda
    if pi is None:
        lrange = sp.linspace(0.05, 0.95, max(lPV / 100.0, 10))
        pil = sp.double((PV[:, sp.newaxis] > lrange).sum(axis=0)) / lPV
        pilr = pil / (1.0 - lrange)
        # ok, I think for SNPs this is pretty useless, pi is close to 1!
        pi = 1.0
        # if there is something useful in there use the something close to 1
        if pilr[-1] < 1.0:
            pi = pilr[-1]

    # 3. initialise q values
    QV_ = pi * m / lPV * PV
    QV_[-1] = min(QV_[-1], 1.0)
    # 4. update estimate
    for i in range(lPV - 2, -1, -1):
        QV_[i] = min(pi * m * PV[i] / (i + 1.0), QV_[i + 1])
    # 5. invert sorting
    QV = sp.zeros_like(PV)
    QV[IPV] = QV_

    QV = QV.reshape(S)
    return QV


def qvalues(pv,
            m=None,
            return_pi0=False,
            lowmem=False,
            pi0=None,
            fix_lambda=None):
    r"""
    FDR estimation using Benjamini-Hochberg and Stories methods

    Args:
        pv (array_like): pvalues.
        m (int): total number of tests. The default value is len(pv).
        return_pi0 (bool): if True, also returns the estimated fraction
                           of null p-values.
        lowmem (bool): employs low-memory implementation that only uses
                       one pv and one qv matrix.
        pi0: fraction of true null.

    Returns:
        (tuple): tuple containing:
            - **qv** (*array_like*): estimated qvalues
            - **pi0** (*float*): estimated fraction of null p-values.
              Only provided if return_counts is True.

    Examples
    --------

        .. doctest::

            >>> from numpy.random import RandomState
            >>> from limix.stats import qvalues
            >>> from numpy import set_printoptions
            >>> set_printoptions(4)
            >>>
            >>> pv = RandomState(1).rand(1000)
            >>> qv, pi0 = qvalues(pv, return_pi0 = True)
            >>> print(qv[:4]) # doctest: +SKIP
            [0.9874 0.9874 0.1144 0.9874]
            >>> print('%.2f' % pi0)
            1.00
    """

    original_shape = pv.shape

    assert (pv.min() >= 0 and pv.max() <= 1)

    pv = pv.ravel(
    )  # flattens the array in place, more efficient than flatten()

    if m is None:
        m = float(len(pv))
    else:
        # the user has supplied an m, let's use it
        m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100:
        pi0 = 1.0
    elif pi0 is not None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])

        if fix_lambda is not None:
            interv_count = (pv > fix_lambda - 0.01).sum()
            uniform_sim = sp.array(
                [(pv > fix_lambda - 0.01).sum() * (i + 1)
                 for i in sp.arange(0, len(sp.arange(0, 0.90, 0.01)))][::-1])
            counts += uniform_sim

        for l in range(len(lam)):
            pi0.append(counts[l] / (m * (1 - lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = sp.interpolate.splrep(lam, pi0, k=3)
        pi0 = sp.interpolate.splev(lam[-1], tck)
        if pi0 > 1:
            LG.warning(
                "got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1"
                % pi0)
            pi0 = 1.0

        assert (pi0 >= 0 and pi0 <= 1), "%f" % pi0

    if lowmem:
        # low memory version, only uses 1 pv and 1 qv matrices
        qv = sp.zeros((len(pv), ))
        last_pv = pv.argmax()
        qv[last_pv] = (pi0 * pv[last_pv] * m) / float(m)
        pv[last_pv] = -sp.inf
        prev_qv = last_pv
        for i in range(int(len(pv)) - 2, -1, -1):
            cur_max = pv.argmax()
            qv_i = (pi0 * m * pv[cur_max] / float(i + 1))
            pv[cur_max] = -sp.inf
            qv_i1 = prev_qv
            qv[cur_max] = min(qv_i, qv_i1)
            prev_qv = qv[cur_max]

    else:
        p_ordered = sp.argsort(pv)
        pv = pv[p_ordered]
        # estimate qvalues
    #     qv = pi0*m*pv/(sp.arange(len(pv))+1.0)

    #     for i in xrange(int(len(qv))-2, 0, -1):
    #         qv[i] = min([qv[i], qv[i+1]])

    qv = pi0 * m / len(pv) * pv
    qv[-1] = min(qv[-1], 1.0)

    for i in range(len(pv) - 2, -1, -1):
        qv[i] = min(pi0 * m * pv[i] / (i + 1.0), qv[i + 1])

    # reorder qvalues
    qv_temp = qv.copy()
    qv = sp.zeros_like(qv)
    qv[p_ordered] = qv_temp

    # reshape qvalues
    qv = qv.reshape(original_shape)

    if return_pi0:
        return qv, pi0
    else:
        return qv


def estimate_lambda(pv):
    """estimate lambda form a set of PV"""
    LOD2 = sp.median(st.chi2.isf(pv, 1))
    L = (LOD2 / 0.456)
    return (L)


def LOD2PV(lods):
    PV = (st.chi2.sf(2 * lods, 1))
    return PV



import sys
import h5py
import pdb
import scipy as SP
import scipy.stats as ST
import scipy.linalg as LA
import time as TIME
import copy
import warnings
import os


class Splitter():

    def __init__(self, pos=None, chrom=None, separate_chroms=False):
        """
        Constructor
        Args:
            pos:        position
            chrom:      chromosome
        """
        # assert
        assert pos is not None, 'Slider:: set pos'
        assert chrom is not None, 'Slider:: set chrom'

        self.pos = pos
        self.chrom = chrom
        # sep chroms
        self.separate_chroms = separate_chroms
        # windows
        self.windows = None
        # additional info for windows
        self.info = {}
        pass

    def splitGeno(
            self,
            method='slidingWindow',
            size=5e4,
            step=None,
            annotation_file=None,
            cis=1e4,
            funct=None,
            minSnps=1.,
            maxSnps=SP.inf,
            cache=False,
            out_dir='./cache',
            fname=None,
            rewrite=False):
        """
        split into windows
        Args:
            method:     method used to slit the windows:
                        'slidingWindow':    uses a sliding window
                        'geneWindow':       uses windows centered on genes
            size:       window size used in slidingWindow method
            step:       moving step used in slidingWindow method
            annotation_file:    file containing the annotation file for geneWindow method
            minSnps:    only windows with nSnps>=minSnps are considered
            maxSnps:    only windows with nSnps>=maxSnps are considered
            cache:      if results need to be cashed
            out_dir:    folder used for caching
            fname:      name of the hdf5 file used for caching
            rewrite:    if true, rewrite the existing cache hdf5 file
        """
        self.info = {}  # forget everything
        # check if it is necessary to read form file or not
        read_from_file = False
        if cache:
            assert fname is not None, 'Splitter:: specify fname'
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            out_file = os.path.join(out_dir, fname)
            read_from_file = os.path.exists(out_file) and not rewrite

        self.size = size

        if read_from_file:
            # reads from file
            f = h5py.File(out_file, 'r')
            self._wnd_pos = f['wnd_pos'][:]
            self._idx_wnd_start = f['idx_wnd_start'][:]
            self._nSnps = f['nSnps'][:]
            f.close()
        else:
            # calculates windows using the indicated method
            if method == 'slidingWindow':
                self._splitGenoSlidingWindow(
                    size=size, step=step, minSnps=minSnps, maxSnps=maxSnps)
            elif method == 'geneWindow':
                self._splitGenoGeneWindow(
                    annotation_file=annotation_file,
                    cis=cis,
                    funct=funct,
                    minSnps=minSnps,
                    maxSnps=maxSnps)

            if cache:
                f = h5py.File(out_file, 'w')
                f.create_dataset('wnd_pos', data=self._wnd_pos)
                f.create_dataset('idx_wnd_start', data=self._idx_wnd_start)
                f.create_dataset('nSnps', data=self._nSnps)
                f.close()

    def _splitGenoSlidingWindow(
            self,
            size=5e4,
            step=None,
            minSnps=1.,
            maxSnps=SP.inf):
        """
        split into windows using a slide criterion
        Args:
            size:       window size
            step:       moving step (default: 0.5*size)
            minSnps:    only windows with nSnps>=minSnps are considered
            maxSnps:    only windows with nSnps>=maxSnps are considered
        """
        if step is None:
            step = 0.5 * size
        chroms = SP.unique(self.chrom)
        wnd_pos = []
        idx_wnd_start = []
        nSnps = []
        wnd_i = 0

        nSnps = []
        for chrom_i in chroms:
            start = 0
            Ichrom = self.chrom == chrom_i
            idx_chrom_start = SP.where(Ichrom)[0][0]
            pos_chr = self.pos[Ichrom]
            pos_chr_max = pos_chr.max()
            while True:
                if start > pos_chr_max:
                    break
                end = start + size
                Ir = (self.pos >= start) * (self.pos < end)
                _nSnps = Ir.sum()
                if _nSnps > minSnps and _nSnps < maxSnps:
                    wnd_pos.append([chrom_i, start, start + size])
                    nSnps.append(_nSnps)
                    idx_wnd_start.append(idx_chrom_start + SP.where(Ir)[0][0])
                    wnd_i += 1
                start += step
        self._wnd_pos = SP.array(wnd_pos)
        self._idx_wnd_start = SP.array(idx_wnd_start)
        self._nSnps = SP.array(nSnps)

    def _splitGenoGeneWindow(
            self,
            annotation_file=None,
            cis=1e4,
            funct='protein_coding',
            minSnps=1.,
            maxSnps=SP.inf):
        """
        split into windows based on genes
        """
        # 1. load annotation
        assert annotation_file is not None, 'Splitter:: specify annotation file'
        try:
            f = h5py.File(annotation_file, 'r')
            geneID = f['geneID'][:]
            gene_chrom = f['chrom'][:]
            gene_start = f['start'][:]
            gene_end = f['end'][:]
            gene_strand = f['strand'][:]
            gene_function = f['function'][:]
            f.close()
        except BaseException:
            print('Splitter:: format annotation file not valid')

        # if funct is not None, it has to be a list
        if funct is not None and funct != list:
            funct = [funct]

        windows = []
        nSnps = []
        Igene = []
        # 2. calculates windows
        for gene_i in range(geneID.shape[0]):
            if funct is not None:
                if gene_function[gene_i] not in funct:
                    Igene.append(False)
                    continue
            wnd = [
                gene_chrom[gene_i],
                gene_start[gene_i] - cis,
                gene_end[gene_i] + cis]
            Ir = (self.chrom == wnd[0]) * \
                (self.pos >= wnd[1]) * (self.pos <= wnd[2])
            _nSnps = Ir.sum()
            if _nSnps >= minSnps and _nSnps <= maxSnps:
                windows.append(wnd)
                nSnps.append(_nSnps)
                Igene.append(True)
            else:
                Igene.append(False)
        Igene = SP.array(Igene)
        self.info['nSnps'] = SP.array(nSnps)
        self.info['geneID'] = geneID[Igene]
        self.info['gene_start'] = gene_start[Igene]
        self.info['gene_end'] = gene_end[Igene]
        self.info['gene_strand'] = gene_strand[Igene]
        self.info['gene_function'] = gene_function[Igene]
        return SP.array(windows)

    @property
    def wnd_pos(self):
        return self._wnd_pos

    @property
    def idx_wnd_start(self):
        return self._idx_wnd_start

    @property
    def nSnps(self):
        return self._nSnps

    @property
    def nWindows(self):
        return self.wnd_pos.shape[0]



import os
import pdb
import unittest

import scipy as sp
import scipy.linalg as LA

from limix.mtset import MTSet

dir_name = os.path.dirname(os.path.realpath(__file__))
base_folder = os.path.join(dir_name, 'data')


class TestMTSet(unittest.TestCase):
    """test class for optimization"""

    def setUp(self):
        sp.random.seed(0)
        self.Y = sp.loadtxt(os.path.join(base_folder, 'Y.txt'))
        self.XX = sp.loadtxt(os.path.join(base_folder, 'XX.txt'))
        self.Xr = sp.loadtxt(os.path.join(base_folder, 'Xr.txt'))
        self.N, self.P = self.Y.shape
        self.write = False

    def test_mtSetNull(self):
        fbasename = 'mtSetNull'
        setTest = MTSet(Y=self.Y, R=self.XX)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg': nullMTInfo['Cg'], 'Cn': nullMTInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSetNull_eigenCache(self):
        fbasename = 'mtSetNull'
        S, U = LA.eigh(self.XX)
        setTest = MTSet(Y=self.Y, S_R=S, U_R=U)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg': nullMTInfo['Cg'], 'Cn': nullMTInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSet_base(self):
        fbasename = 'mtSet_base'
        setTest = MTSet(self.Y, R=self.XX)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr': optInfo['Cr'], 'Cg': optInfo['Cg'], 'Cn': optInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSet_eigenCache(self):
        fbasename = 'mtSet_base'
        S, U = LA.eigh(self.XX)
        setTest = MTSet(Y=self.Y, S_R=S, U_R=U)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr': optInfo['Cr'], 'Cg': optInfo['Cg'], 'Cn': optInfo['Cn']}
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSetPCnull(self):
        fbasename = 'mtSetPCnull'
        setTest = MTSet(Y=self.Y)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cn': nullMTInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSetPCnull_fixed(self):
        fbasename = 'mtSetPCnull_fixed'
        setTest = MTSet(Y=self.Y, F=self.Xr)
        nullMTInfo = setTest.fitNull(cache=False)
        ext = {'Cg': nullMTInfo['Cg'], 'Cn': nullMTInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSetPC_base(self):
        fbasename = 'mtSetPC_base'
        setTest = MTSet(Y=self.Y)
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr': optInfo['Cr'], 'Cn': optInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def test_mtSetPC_fixed(self):
        fbasename = 'mtSetPC_fixed'
        setTest = MTSet(Y=self.Y, F=self.Xr[:, :2])
        optInfo = setTest.optimize(self.Xr)
        ext = {'Cr': optInfo['Cr'], 'Cn': optInfo['Cn']}
        if self.write:
            self.saveStuff(fbasename, ext)
        RV = self.assess(fbasename, ext)
        self.assertTrue(RV)

    def saveStuff(self, fbasename, ext):
        """ util function """
        base = os.path.join(base_folder, 'res_' + fbasename + '_')
        for key in list(ext.keys()):
            sp.savetxt(base + key + '.txt', ext[key])

    def loadStuff(self, fbasename, keys):
        """ util function """
        RV = {}
        base = os.path.join(base_folder, 'res_' + fbasename + '_')
        for key in keys:
            RV[key] = sp.loadtxt(base + key + '.txt')
        return RV

    def assess(self, fbasename, ext):
        """ returns a bool vector """
        real = self.loadStuff(fbasename, list(ext.keys()))
        RV = sp.all([((ext[key] - real[key])**2).mean() < 1e-6
                     for key in list(ext.keys())])
        if not RV:
            pdb.set_trace()
        return RV


if __name__ == '__main__':
    unittest.main()
